#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


"""
Postfix to DTN Interface - main routine

This code sets up the set of threads that manage the interface:

The interface manages five things:
- Managing outgoing messages coming from Postfix and sending them to dtnd
- Managing incoming messages coming from dtnd and sending them to Postfix
- Maintaining a maildir that contains copies of all mail messages sent out
  to dtnd but for which no delivery confirmation has yet been received.
- Maintaining a maildir that contains copies of  messages received from dtnd
  that it has not yet been possible to relay to Postfix
- Managing timers that determine when a message needs to be resent because it
  will have expired in the network subject to limit on the number of retries.
  Messages that have not been delivered after the retry limit are deleted
  and an 'undeliverable' message sent to the sender via Postfix.
- Managing a list of recently received messages that allows duplicate messages
  received from dtnd to be suppressed. The list is purged at  the expiry time
  of the bundle that brought the first copy plus a grace period.


There are four permanent threads created:
- Thread to accept connection requests from Postfix outbound pipe shim
    This thread opens a listening socket and waits for (local) connections
    from the Postfix outbound pipe shim.  Whenever a new connection is made
    Postfix will be relaying a (single) message.  This is expected to be in
    RFC 2822 format as with BSMTP (i.e.lines terminated with \r\n, lines
    containing only a . being turned into ..) and with a Delibvered To:
    header prepended.  The listener thread spawns a new thread that reads
    the message and dumps it into a file with a unique name in the maildir.
    When the message has been completly received a send request is queued
    for the the dtn manager thread.
- Thread to handle the dtnd connection.  The dtnd connection can both send
  and receive bundles in a sort of half duplex arrangement.  The thread
  manages the connection, waiting for incoming bundles when there is
  nothing else to do, periodically checking the send request queue and sending
  any messages in the request queue as bundles.  Incoming bundles with messages
  are dumped into files in the maildir (or if received as files, moved there).
  The file format is checked and the Delivered To: header checked. The database
  of recently received messages is checked to detect duplicates and duplicates
  discarded. If the header is for the DTN association being managed, the message
  is queued for resending. Otherwise a request to send to Postfix is queued.
  Bundles may also contain delivery reports. The mail message associated with
  the report is deleted from the outgoing maildir and records of the message
  are discarded.
- Thread to manage sending incoming messages (and delivery failure reports).
  This can be done either by calling sendmail or making an SMTP connection.
  Optimum solution TBD.  Probably SMTP.  Requests are received from a queue.
  Successfully sent mails are deleted from the maildir but the message is
  recorded in a database to manage duplicate suppression.
- Thread to handle manage message storage.  Deals with resends and sending
  delivery failure reports.  Queues requests on the dtnd and Postfix dending
  threads as required. On startup checks for unsent messages (either way)
  and requeues them for sending.

Two maildirs are used: one each for incoming and outgoing mail.

The records of timing and duplicate testing are (will be) held in a SQLite
database

NOTE:
The Postfix outging mail server threads make use of the Python SocketServer
framework. Frankly, the code in the Python 2.5 library is complete rubbish.
It is esentially impossible to set the reuse address flag making (at least) testing
very difficult.  Further it is extremely difficult to shutdown the servers
cleanly.  Most of this is fixed in the Python 2.6 release. Since this module
is pure Python and doesn't do anything v2.6 specific itself, I have extracted
the module and am using for this program.
"
***************************************
This version can be configured to deal with either direct to Nomadic stations
or using the NAT relay that is reached with static routing before transitioning
to dynamic routing for the final stage. This just involves setting the boolean
flag relay_mode when creating the dtn_send thread.
***************************************

Revision History
================
Version   Date       Author         Notes
0.1       14/03/2010 Elwyn Davies   License statement added.
                                    Relay mode switch incorporated.
0.0	  25/07/2009 Elwyn Davies   Created for N4C Summer tests 2009
"""

import os
import time
import threading
import Queue
import socket
from SocketServer2_6 import *
import logging
import logging.handlers
import logging.config

import pypop_maildir
import dp_pfmailout
import dp_pfmailin
import dp_dtn
import dtnapi
from dp_msg_send_evt import *

def main(mailbox_root, log_file, log_config_file):
    # Setup logging
    logging.config.fileConfig(log_config_file)
    pymail_logger = logging.getLogger("dtn_pymail")
    loginfo = pymail_logger.info
    logdebug = pymail_logger.debug

    # Set domain for which we are handling mail
    mail_domain = "nomadic.n4c.eu"
    
    # Check mailboxes dtnin and dtnout  exist and create if not ***TBD***
    nomadic_domain = pypop_maildir.MaildirDirdbmDomain(mailbox_root)
    if not nomadic_domain.userDirectory("dtnin"):
        nomadic_domain.addUser("dtnin", "password")
    if not nomadic_domain.userDirectory("dtnout"):
        nomadic_domain.addUser("dtnout", "password")
    dtnin_mailbox  = pypop_maildir.MaildirMailbox(nomadic_domain.userDirectory("dtnin"),
                                                  "dtnin")
    dtnout_mailbox = pypop_maildir.MaildirMailbox(nomadic_domain.userDirectory("dtnout"),
                                                  "dtnout")

    # Create queues to send requests for messages to be written out
    # FIFO queues with no depth limitation at present
    dtn_send_q = Queue.Queue()
    postfix_send_q = Queue.Queue()

    #====================================================================#
    # Create server to handle mails from Postfix
    pfmailout = dp_pfmailout.postfix_mailout_server(dtnout_mailbox, dtn_send_q,
                                                    pymail_logger)

    # Start a thread with the server -- that thread will then start one
    # more thread for each request
    mailout_listener = threading.Thread(target=pfmailout.serve_forever)
    # Exit the server thread when the main thread terminates
    mailout_listener.setDaemon(True)
    mailout_listener.start()
    
    #====================================================================#
    # Create threads to handle DTN interface
    # There are three threads that desl with:
    # - sending emails
    # - receiving emails
    # - receiving reports

    # Sending thread
    dtn_send_handler = dp_dtn.dtn_send(mail_domain,
                                       dtn_send_q, pymail_logger, 
				       relay_mode=True)
    dtn_send_handler.setDaemon(True)
    dtn_send_handler.start()

     # Receiving thread
    dtn_receive_handler = dp_dtn.dtn_receive(mail_domain, dtnin_mailbox,
                                             postfix_send_q, pymail_logger)
    dtn_receive_handler.setDaemon(True)
    dtn_receive_handler.start()
    
    # Report handler thread
    dtn_report_handler = dp_dtn.dtn_report(mail_domain, pymail_logger)
    dtn_report_handler.setDaemon(True)
    dtn_report_handler.start()
    
    #====================================================================#
    # Sender for incoming mails to  be deleivered by Postfix
    postfix_sender = dp_pfmailin.PostfixMailinHandler(mail_domain, dtnin_mailbox,
                                                    postfix_send_q, pymail_logger)
    postfix_sender.setDaemon(True)
    postfix_sender.start()
 

    # Let everything have a chance to get going (old trick)
    time.sleep(0.1)
    loginfo("Postfix outgoing mail listener running in thread: %s" % mailout_listener.getName())
    loginfo("Postfix incoming mail sender running in thread: %s" % postfix_sender.getName())
    loginfo("DTN Send handler running in thread: %s" % dtn_send_handler.getName())
    loginfo("DTN Receive handler running in thread: %s" % dtn_receive_handler.getName())

    loginfo("DTN Report handler running in thread: %s" % dtn_report_handler.getName())   
      
    try:
        while True:
            pass
    except:
        # Shutdown
        # Think a bit about ordering here...
        # Shutdown the producer threads first
        dtn_receive_handler.end_run()
        pfmailout.end_run()

        # Give them a few seconds to die...
        time.sleep(6)

        # Then shutdown the consumer threads
        dtn_report_handler.end_run()
        evt = msg_send_evt(MSG_END,
                           "",
                           MSG_NEW,
                           "")
        dtn_send_q.put_nowait(evt)
        postfix_send_q.put_nowait(evt)

        # ..and wait for them to die...
        time.sleep(2)                                    
        

if __name__ == "__main__":
    main("/home/dtn/maildir", "/home/dtn/log/dtn_pymail.log", "/home/dtn/dtn/dtn_pymail_log.conf")
    # Test code

