#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


"""
DTN-Postfix interface

Event object class

Instances of this object are passed between the various threads of the
system in the send queues to tell the sending threads what work they have
to do.

Each contains
- the filename of a file that contains the email to send
- a flag indicating if the item is new or a resend
- (for debugging purposes) an indicator of whether DTN or Postfix send is needed

Revision History
================
Version   Date       Author         Notes
0.1       14/03/2010 Elwyn Davies   License statement added.
0.0	  24/07/2009 Elwyn Davies   Created for N4C Summer tests 2009
"""
import os

# Whether the message is new or a retry (affects what we do to the database)
MSG_NEW = 0
MSG_RETRY = 1

# Type send needed - or terminate sending process
MSG_DTN = "dtn_send"
MSG_SMTP = "smtp_send"
MSG_END = "end_send"


class msg_send_evt:
    def __init__(self, send_type, filename, new_or_retry, destination):

        # Check the parameters are valid
        # (note the file *might* change under our feet - no guarantees later)
        if send_type not in (MSG_DTN, MSG_SMTP, MSG_END):
            raise ValueError
        if (send_type != MSG_END):
            if new_or_retry not in (MSG_NEW, MSG_RETRY):
                raise ValueError
            if not os.access(filename, os.R_OK):
                raise ValueError
        else:
            filename = ""
            new_or_retry = MSG_NEW
            destination = ""
        
        self._filename =  filename
        self.new_or_retry = new_or_retry
        self.send_type = send_type
        self._destination = destination

    def is_new(self):
        return self.new_or_retry == MSG_NEW

    def is_last_msg(self):
        return self.send_type == MSG_END

    def filename(self):
        return self._filename

    def destination(self):
        return self._destination

    def __repr__(self):
        if self.send_type == MSG_END:
            return "Ending sending."
        else:
            return "%s message in file %s to be sent by %s to %s" % ({ MSG_NEW: "New", MSG_RETRY: "Resending"}[self.new_or_retry],
                                                                     self._filename,
                                                                     self.send_type,
                                                                     self._destination)

if __name__ == "__main__":
    evt = msg_send_evt(MSG_DTN, "/etc/passwd", MSG_RETRY, "abc@def.ghi")
    print evt
    evt = msg_send_evt(MSG_END, "", MSG_NEW, "")
    print evt
    evt = msg_send_evt(MSG_SMTP, "/etc/passwd", 3, "abd@def.ghi")
