#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

"""
DTN-Postfix interface

Handlers for messages being sent to Postfix after being relayed through DTN

The thread makes a connection to Postfix over local SMTP and sends messages
using the smtplib.  The messages are not significantly processed, but
it is necessary to get the To and From addresses in order to build a wrapper.

The message is stored in a maildir file  - the lines are terminated with \n

Revision History
================
Version   Date       Author         Notes
0.1       14/03/2010 Elwyn Davies   License statement added.
0.0	  24/07/2009 Elwyn Davies   Created for N4C Summer tests 2009
"""

import os
import socket
import threading
import smtplib
import email
import logging

import Queue

import pypop_maildir
from dp_msg_send_evt import *

# Special email header carrying recipient addresses across DTN
DTN_RCPTS_HDR = "X-DTN-rcpts"

class PostfixMailinHandler(threading.Thread):

    def __init__(self, domain, mailbox, send_q, logger):
        threading.Thread.__init__(self, name="postfix-send")
        self.domain = domain
        self.mailbox = mailbox
        self.send_q = send_q
        self.logger = logger

        # Flag to keep the loop running
        self.mailin_run = True

    def end_run(self):
        self.mailin_run = False
        
    def run(self):
        self.mailin_thread = threading.currentThread()

        # Logging functions
        self.loginfo = self.logger.info
        self.logdebug = self.logger.debug
        self.logwarn = self.logger.warn
        self.logerror = self.logger.error

        # Create an SMTP connection manager for the mailserver on localhost
        smtp_link = smtplib.SMTP()

        # Loop waiting for messages to come from DTN (or end message)
        self.loginfo("Entering SMTP send Queue loop")
        
        # The thread sits waiting for queued events forever.
        # The thread is terminated by sending a special MSG_END event
        while (self.mailin_run):
            evt = self.send_q.get()
            if evt.is_last_msg():
                self.loginfo("Mailin sender finishing")
                break

            # The received evt points to a file containing the email
            # message to be sent
            # Open it and parse the message to make sure it is real
            self.logdebug("new mail send event received")
            try:
                fp = None
                fp = open(evt.filename(), "r")
                email_msg = email.message_from_file(fp)
                fp.close()

            except:
                print "Unable to extract mail message from %s." % evt.filename()
                if fp != None and not fp.closed:
                    fp.close()

                # We should trash this mail message and possibly send a non-delivery
                # message.. but if the message is seriously garbled there won't be
                # anything we can do.  Don't do anything for the time being
                # We should move this out of the new folder at least.
                continue

            # The message ought to contain the special X- header with the reciepients
            # that were given to the SMTP daemon on the other side of the link
            if email_msg.has_key(DTN_RCPTS_HDR):
                rcpts_str = email_msg.get(DTN_RCPTS_HDR)
            else:
                self.logerror("Email message in %s from DTN doesn't have %s header" % 
                              (evt.filename(), DTN_RCPTS_HDR))

                # Again should get rid of these message
                continue

            # Extract the addresses from the value string
            self.logdebug("Sending to %s" % rcpts_str)
            smtp_rcpts = rcpts_str.split()
            rcpts = []
            for r in smtp_rcpts:
                if r[0] == "<" and r[-1] == ">" and r != "<>":
                    rcpts.append(r[1:-1])

            # The message should have a Unix From envelope header
            unixfrom = email_msg.get_unixfrom()
            if unixfrom == None:
                self.logerror("Email message in %s from DTN doesn't have From envelope header" %
                              evt.filename())

                # Again get rid of the message
                continue
            else:
                from_cmpts = unixfrom.split()
                if len(from_cmpts) < 2:
                    self.logerror("Email message in %s from DTN has malformed From envelope header" %
                                  evt.filename())

                    # Again get rid of the message
                    continue
                else:
                    from_addr = from_cmpts[1]

            # We can now send the email to Postfix (on localhost - defaults
            # for connect are port 25 on localhost)
            try:
                self.logdebug("Opening link to Postfix")
                smtp_link.connect()
                smtp_link.sendmail(from_addr, smtp_rcpts, email_msg.as_string(True))
                self.loginfo("Email in %s passed to Postfix for sending to %s" %
                             (evt.filename(), " ".join(smtp_rcpts)))
            except:
                self.logerror("Unable to send message in %s to Postfix." %
                              evt.filename())
            finally:
                self.logdebug("Closing link to Postfix")
                smtp_link.quit()
if __name__ == "__main__":
    import time
    logger = logging.getLogger("test")
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    fmt = logging.Formatter("%(levelname)s %(threadName)s %(message)s")
    ch.setFormatter(fmt)
    logger.addHandler(ch)

    postfix_send_q = Queue.Queue()

    # Set domain for which we are handling mail
    mail_domain = "nomadic.n4c.eu"
    
    # Check mailboxes dtnin and dtnout  exist and create if not ***TBD***
    nomadic_domain = pypop_maildir.MaildirDirdbmDomain("/tmp")
    if not nomadic_domain.userDirectory("dtnin"):
        nomadic_domain.addUser("dtnin", "password")
    dtnin_mailbox  = pypop_maildir.MaildirMailbox(nomadic_domain.userDirectory("dtnin"),
                                                  "dtnin")
    server = PostfixMailinHandler(mail_domain, dtnin_mailbox,
                                  postfix_send_q, logger)
    server.setDaemon(True)
    server.start()
    time.sleep(0.1)
    logger.info("Postfix mail sender running in thread: %s" % server.getName())

    evt = msg_send_evt(MSG_SMTP,
                       "/home/elwynd/Python/pypop/msg3",
                       MSG_NEW,
                       ["elwynd@mightyatom.foly.org.uk"])
    postfix_send_q.put_nowait(evt)

    time.sleep(1)

    evt = msg_send_evt(MSG_END,
                       "",
                       MSG_NEW,
                       "")
    postfix_send_q.put_nowait(evt)

    # ..and wait for them to die...
    time.sleep(2)                                    

    

    
    

    
