#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


"""
DTN-Postfix interface

Handlers for messages sent from Postfix being relayed through DTN

The script run by the pipe daemon makes a local  TCP connection to the
PostfixMailoutServer (subclassed from SoCketServer's threaded TCP Server)
Each new mail from Postfix creates a new thread to receive the mail and
dump it into a file in the dtnout maildir.

The 'protocol' expects a single RFC822 message per PostfixMailoutHandler
invocation with a Delivered-To header on the front. This will be preserved
in the file. The address is used to create the detination EID later.

The message is stored in a maildir file  - the lines are terminated with \n

Revision History
================
Version   Date       Author         Notes
0.1       14/03/2010 Elwyn Davies   License statement added.
0.0	  24/07/2009 Elwyn Davies   Created for N4C Summer tests 2009
"""

import os
import socket
from SocketServer2_6 import *
import threading
import logging
import Queue

import pypop_maildir
import dp_msg_send_evt

class PostfixMailoutHandler(StreamRequestHandler):

    lineseps = "\r\n"
    del_to = "Delivered-To: "
    del_to_len = len(del_to)

    def end_run(self):
        self.request.close()

    def handle(self):
        self.mailout_thread = threading.currentThread()
        self.mailout_thread.setName("Mailout receiver - %d" %
                                    self.server.next_server_num)
        self.server.next_server_num += 1

        # Tell listener we are running
        self.server.add_thread(self)
        
        # Logging functions
        self.loginfo = self.server.logger.info
        self.logdebug = self.server.logger.debug
        self.logwarn = self.server.logger.warn
        self.logerror = self.server.logger.error

        self.loginfo("New mailout connection from %s" % self.client_address[0])

        # The first line should start "Delivered-To: "
        # The rest of the line is the target destination address
        # We use this to determine the EID that is used to send
        # this email across the DTN
        data = self.rfile.readline().rstrip(self.lineseps)
        if not data.startswith(self.del_to):
            # Protocol error
            self.logerror("Message does not start with 'Delivered-To:'")
            return
        destn = data[self.del_to_len:].strip()
        self.loginfo("Receiving message for delivery to %s" % destn )

        # Initialize message destined for 'dest' in given mailbox
        msg = self.server.mailbox.newMessage(destn)
        self.logdebug("ready to receive message")

        try:
            while True:
                data = self.rfile.readline()
                if data == "":
                    break
		data = data.rstrip(self.lineseps)
                self.logdebug(data)
                msg.lineReceived(data)
        except:
            msg.connectionLost()
            self.request.close()
            return
        
        # Finalize message and put it where it belongs
        self.logdebug("Finishing message")
        if msg.eomReceived():
            fn = msg.getFinalName()

            # Queue the message for sending by DTN
            self.loginfo("Message to be sent by DTN stored as %s" % fn )
            try:
                evt = dp_msg_send_evt.msg_send_evt(dp_msg_send_evt.MSG_DTN,
                                                   fn,
                                                   dp_msg_send_evt.MSG_NEW,
                                                   [destn])
                self.server.send_queue.put_nowait(evt)
            except Queue.Full:
                # The periodic thread will pick up the new message later
                self.logdebug("Send queue was full")
                pass
        self.loginfo("Mailout receiver finishing")
        return
        

class PostfixMailoutServer(ThreadingMixIn, TCPServer):
    def __init__(self, addr, handler, mailbox, send_queue, logger):
        # These are used  by individual requests
        # accessed via self.server in the handle function
        self.mailbox = mailbox
        self.send_queue = send_queue
        self.logger = logger
        
        self.running_threads = set()
        self.next_server_num = 1

        # Setup to produce a daemon thread for each incoming request
        # and be able to reuse address
        TCPServer.__init__(self, addr, handler,
                           bind_and_activate=False)
        self.allow_reuse_address = True
        self.server_bind()
        self.server_activate()
                         
        self.daemon_threads = True

    def add_thread(self, thread):
        self.logger.debug("New thread added")
        self.running_threads.add(thread)

    def remove_thread(self, thread):
        if thread in self.running_threads:
            self.running_threads.remove(thread)

    def end_run(self):
        for thread in self.running_threads:
            if thread.mailout_thread.isAlive():
                thread.request.close()
        del self.running_threads
        self.shutdown()

def postfix_mailout_server(mailbox, send_queue, logger):
    HOST, PORT = "localhost", 2111
    return PostfixMailoutServer((HOST, PORT), PostfixMailoutHandler,
                                mailbox, send_queue, logger)



if __name__ == "__main__":
    import time
    logger = logging.getLogger("test")
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    fmt = logging.Formatter("%(levelname)s %(threadName)s %(message)s")
    ch.setFormatter(fmt)
    logger.addHandler(ch)

    nomadic_domain = "nomadic.n4c.eu"
    tmp_domain = pypop_maildir.MaildirDirdbmDomain("/tmp")
    if not tmp_domain.userDirectory("dtnout"):
        tmp_domain.addUser("dtnout", "password")
    dtnout_mailbox = pypop_maildir.MaildirMailbox(tmp_domain.userDirectory("dtnout"),
                                                 "dtnout")
    
    send_q = Queue.Queue()
                          
    def client(ip, port, message):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((ip, port))
        sock.send(message)
        #response = sock.recv(1024)
        #print "Received: %s" % response
        sock.close()

    # Port 0 means to select an arbitrary unused port
    HOST, PORT = "localhost", 0

    server = PostfixMailoutServer((HOST, PORT), PostfixMailoutHandler,
                                  dtnout_mailbox, send_q, logger)
    ip, port = server.server_address

    # Start a thread with the server -- that thread will then start one
    # more thread for each request
    server_thread = threading.Thread(target=server.serve_forever,
                                     name="Mailout Listener")
    # Exit the server thread when the main thread terminates
    server_thread.setDaemon(True)
    server_thread.start()
    logger.info("Server loop running in thread:%s" % server_thread.getName())

    client(ip, port, "Delivered-To: a@b.c\r\nmessage\r\n")
    #client(ip, port, "Delivered-To: d@b.c\r\nmessage\r\nHello World 2\r\n")
    #client(ip, port, "Delivered-To: ae@b.c\r\nmessage\r\nHello World 3\r\n")
                
    time.sleep(1)
    while not send_q.empty():
        evt = send_q.get()
        logger.info("Msg to %s" % evt.destination())

    time.sleep(1)

    server.end_run()
