#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


"""
Maildir-style mailbox support

Intended ultimately for use on Symbian-based mobile pnones in conjunction
with DTN package.

This is intended to be a very lightweight system primaruily accessed only
on the local machine.   We need a system for storing mail messages to be received
from DTN so that they can be accessed by the standard phone messaging system.

Standard email on a mobile phone doesn't readily handle bulk storage of emails
(as opposed to text messages etc).  The intention is that these mailboxes are
accessed by a lightweight POP3 protocol and is adapted as such.

Mailboxes are associated with:
- a (mail) domain
- a directory structure
A domain has a root directory which contains:
- a file called 'passwd' which hold the passwords for all the users
- a directory for each user named by their username (avatar id)
Each user directory contains the usual set of maildir subdirectories
tmp, new, cur and .trash.  The maildir operates by means of the 'standard'
maildir semantics.

When the system is active, a user's mailbox in the domain is represented by
an 'avatar' which is a MaildirMailbox object instance.  The domain maintains
one instance for each active user.  It is intended that there should only be
one client at a time for each user, but the system should work OK with multiple
clients subject to some thought about thread safeness.  However, the nature
of a mobile phone makes this an unlikely situation and we won't worry too hard
about it at the moment.

The domain supports authentication via username/password (plain text) and APOP
style hashed password.  No great effort (actually no effort at all)is made to
keep the passwords securely since the intention is that the maildir will live
on the phone where it is accessed and the files can be read by anybody who is
able to access the phone memory.  Usernames are desirable so that a phone can
be used for multiple incoming mail addresses, but a phone is inherently a single
user device but normally the communication between mail client and mailbox will
be over a channel within the phone so that the user name and password are never
visible outside the phone.  On the other hand, one might have to give a bit more
thought to this if the system was used to copy mail off the phone via a POP3
mail client on another device.


Revision History
================
Version   Date       Author         Notes
0.1       14/03/2010 Elwyn Davies   License statement added.
0.0       01/06/2009 Elwyn Davies   Initial version based on ideas from the
                                    Twisted Python implementation.
"""

import os
import stat
import socket
import time
import hashlib

try:
    import cStringIO as StringIO
except ImportError:
    import StringIO

try:
    import e32dbm as anydbm
except ImportError:
    import anydbm

class NotImplementedError(Exception):
    pass

class UnauthorizedLogin(Exception):
    pass

class _MaildirNameGenerator:
    """
    Utility class to generate a unique maildir name

    """
    n = 0
    p = os.getpid()
    hn = socket.getfqdn()

    def __init__(self):
        return

    def generate(self):
        """
        Return a string which is intended to unique across all calls to this
        function (across all processes, reboots, etc).

        Strings returned by earlier calls to this method will compare less
        than strings returned by later calls as long as the clock provided
        doesn't go backwards.
        """
        self.n = self.n + 1
        t = time.time()
        seconds = int(t)
        microseconds = '%07d' % (int((t - seconds) * 10e6),)
        return '%s.M%sP%sQ%s.%s' % (str(seconds), microseconds,
                                    self.p, self.n, self.hn)

_generateMaildirName = _MaildirNameGenerator().generate

def initializeMaildir(dir):
    if not os.path.isdir(dir):
        os.mkdir(dir, 0700)
        for subdir in ['new', 'cur', 'tmp', '.Trash']:
            os.mkdir(os.path.join(dir, subdir), 0700)
        for subdir in ['new', 'cur', 'tmp']:
            os.mkdir(os.path.join(dir, '.Trash', subdir), 0700)
        # touch
        open(os.path.join(dir, '.Trash', 'maildirfolder'), 'w').close()

class MaildirMessage:
    size = None
    UNIXFROM = "Fron "
    UNIXFROM_LEN = len(UNIXFROM)
    X_DTN_HDR_NAME = "X-DTN-rcpts"

    def __init__(self, address, fp, name, finalName):
        self.fp = fp
        self.name = name
        self.address = address
        self.finalName = finalName
        self.first_line = True
        self.size = 0

    def lineReceived(self, line):
        # Add a X-DTN-rcpts header after a possible From envelope header
        # which must be the first line received
        if self.first_line:
            dtn_header = "%s: %s\n" % (self.X_DTN_HDR_NAME, self.address)
            if line[:self.UNIXFROM_LEN]:
                self.fp.write(line + '\n')
                self.fp.write(dtn_header)
            else:
                self.fp.write(dtn_header)
                self.fp.write(line + '\n')
            self.size += len(dtn_header)
            self.first_line = False
        else:
            self.fp.write(line + '\n')
        self.size += len(line)+1

    def eomReceived(self):
        self.finalName = self.finalName+',S=%d' % self.size
        self.fp.close()
        try:
            os.rename(self.name, self.finalName)
            return True
        except:
            os.remove(self.name)
            return False
        
    def connectionLost(self):
        self.fp.close()
        os.remove(self.name)

    def getFinalName(self):
        return self.finalName
        
class AbstractMaildirDomain:
    """Abstract maildir-backed domain.
    """
    root = None
    _credcheckers = None

    def __init__(self, root):
        """Initialize.
        """
        self.root = root

    def userDirectory(self, user):
        """Get the maildir directory for a given user

        Override to specify where to save mails for users.
        Return None for non-existing users.
        """
        return None

    ##
    ## IDomain
    ##
    def exists(self, user):
        """Check for existence of user in the domain
        """
        if self.userDirectory(user.dest.local) is not None:
            return lambda: self.startMessage(user)
        else:
            raise pypop_smtp.SMTPBadRcpt(user)

    def startMessage(self, destn):
        """
        Save a message for a given user.
        """
        if isinstance(destn, str):
            if destn.find("@") == (-1):
                name = destn
                domain = "localhost"
            else:
                name, domain = destn.split('@', 1)
        else:
            name, domain = destn.dest.local, destn.dest.domain

        dirname = self.userDirectory(name)
        fname = _generateMaildirName()
        filename = os.path.join(dirname, 'tmp', fname)
        fp = open(filename, 'w')
        return MaildirMessage('%s@%s' % (name, domain), fp, filename,
                              os.path.join(dirname, 'new', fname))

    def willRelay(self, user, protocol):
        return False

    def addUser(self, user, password):
        raise NotImplementedError

    def getCredentialsCheckers(self):
        raise NotImplementedError
    ##
    ## end of IDomain
    ##


class MaildirMailbox:
    """Implement the POP3 mailbox semantics for a Maildir mailbox
    
    @type loginDelay: C{int}
    @ivar loginDelay: The number of seconds between allowed logins for the
    user associated with this mailbox.  None

    @type messageExpiration: C{int}
    @ivar messageExpiration: The number of days messages in this mailbox will
    remain on the server before being deleted.
    """
    def __init__(self, path, owner):
        """Initialize with name of the Maildir mailbox
        """
        self.path = path
        self.owner = owner
        self.deleted = {}
        self.list = []
        initializeMaildir(path)
        self.makeMessageList()

    def makeMessageList(self):
        self.list = []
        for name in ('cur', 'new'):
            for file in os.listdir(os.path.join(self.path, name)):
                self.list.append((file, os.path.join(self.path, name, file)))
        self.list.sort()
        self.list = [e[1] for e in self.list]

    def listMessages(self, i=None):
        """Retrieve the size of one or more messages.

        @type index: C{int} or C{None}
        @param index: The number of the message for which to retrieve the
        size (starting at 0), or None to retrieve the size of all messages.

        @rtype: C{int} or any iterable of C{int} or a L{Deferred} which fires
        with one of these.

        @return: The number of octets in the specified message, or an iterable
        of integers representing the number of octets in all the messages.  Any
        value which would have referred to a deleted message should be set to 0.

        @raise ValueError: if C{index} is greater than the index of any message
        in the mailbox.
        
        Return a list of lengths of all files in new/ and cur/
        """
        if i is None:
            ret = []
            for mess in self.list:
                if mess:
                    ret.append(os.stat(mess)[stat.ST_SIZE])
                else:
                    ret.append(0)
            return ret
        else:
            if (i < 0) or (i >= len(self.list)):
                raise ValueError
            else:
                return self.list[i] and os.stat(self.list[i])[stat.ST_SIZE] or 0

    def getMessage(self, i):
        """Retrieve a file-like object for a particular message.

        @type index: C{int}
        @param index: The number of the message to retrieve

        @rtype: A file-like object
        @return: A file containing the message data with lines delimited by
        C{\\n}.
        
        Return an open file-pointer to a message
        """
        if (i < 0) or (i >= len(self.list)) or (self.list[i] == 0):
            raise ValueError
        return open(self.list[i])

    def getUidl(self, i):
        """Get a unique identifier for a particular message.

        @type index: C{int}
        @param index: The number of the message for which to retrieve a UIDL

        @rtype: C{str}
        @return: A string of printable characters uniquely identifying for all
        time the specified message.

        @raise ValueError: if C{index} is greater than the index of any message
        in the mailbox.
        
        Return a unique identifier for a message

        This is done using the basename of the filename.
        It is globally unique because this is how Maildirs are designed.
        """
        # Returning the actual filename is a mistake.  Hash it.
        if (i < 0) or (i >= len(self.list)) or (self.list[i] == 0):
            raise ValueError
        base = os.path.basename(self.list[i])
        m = hashlib.md5()
        m.update(base)
        return m.hexdigest()

    def deleteMessage(self, i):
        """Delete a particular message.

        This must not change the number of messages in this mailbox.  Further
        requests for the size of deleted messages should return 0.  Further
        requests for the message itself may raise an exception.

        @type index: C{int}
        @param index: The number of the message to delete.
        
        Delete a message

        This only moves a message to the .Trash/ subfolder,
        so it can be undeleted by an administrator.
        """
        if (i < 0) or (i >= len(self.list)) or (self.list[i] == 0):
            raise ValueError
        
        trashFile = os.path.join(
            self.path, '.Trash', 'cur', os.path.basename(self.list[i])
        )
        os.rename(self.list[i], trashFile)
        self.deleted[self.list[i]] = (i, trashFile)
        self.list[i] = 0

    def undeleteMessages(self):
        """
        Undelete any messages which have been marked for deletion since the
        most recent L{sync} call.

        Any message which can be undeleted should be returned to its
        original position in the message sequence and retain its original
        UID.
        
        Undelete any deleted messages it is possible to undelete

        This moves any messages from .Trash/ subfolder back to their
        original position, and empties out the deleted dictionary.
        """
        for (real, trash) in self.deleted.items():
            try:
                os.rename(trash[1], real)
            except OSError, (err, estr):
                import errno
                # If the file has been deleted from disk, oh well!
                if err != errno.ENOENT:
                    raise
                # This is a pass
            else:
                try:
                    self.list[trash[0]] = real
                except ValueError:
                    self.list.append(real)
        self.deleted.clear()

    def moveNewCur(self, i=None):
        """
        Move one or all messages in new to cur (not including any that
        might have been deleted).
        """
        newdir = os.path.join(self.path, "new")
        if i == None:
            i = 0
            for m in self.list:
                if m != 0:
                    (pathname, base) = os.path.split(m)
                    if pathname == newdir:
                        n = os.path.join(self.path, "cur", base)
                        os.rename(m, n)
                        self.list[i] = n
                i +=1
        else:
            if (i < 0) or (i >= len(self.list)) or (self.list[i] == 0):
                raise ValueError
            
            m = self.list[i]
            if m != 0:
                (pathname, base) = os.path.split(m)
                if pathname == newdir:
                    n = os.path.join(self.path, "cur", base)
                    os.rename(m, n)
                    self.list[i] = n
    def cleanTrash(self):
        """
        Remove any messages in deleted and compress the list.
        When using POP3 to access the mailbox this will be done
        in the UPDATE state when QUIT is called.  The list is kept in step
        so that the instance can be used with a new session.
        """
        for (real, trash) in self.deleted.items():
            try:
                os.remove(trash[1])
            except:
                pass
            else:
                del self.list[trash[0]]

    def newMessage(self, destn):
        fname = _generateMaildirName()
        filename = os.path.join(self.path, 'tmp', fname)
        fp = open(filename, 'w')
        return MaildirMessage(destn, fp, filename,
                              os.path.join(self.path, 'new', fname))

    def newMessageFile(self, temp_fname):
        fname = _generateMaildirName()
        try:
            size = os.stat(temp_fname)[stat.ST_SIZE]
            new_fn = os.path.join(self.path, 'new', "%s,S=%d" % (fname, size) )
            os.rename(temp_fname, new_fn)
        except:
            return None
        self.list.append(new_fn)
        return new_fn
        
        
    def appendMessage(self, msg):
        """
        Appends a message into the mailbox.
        called when a new message has been installed dynamically
        to add to the list.

        This is an optimization to avoid rereading the files. 
        """
        self.list.append(msg.getFinalName())

    def sync(self):
        """
        No specific requirements.
        """
        pass

##
## Authentication
##
class CredentialsChecker:
    def checkPassword(self, password):
        raise NotImplementedError

class UsernameHashedPassword(CredentialsChecker):

    def __init__(self, username, hashed):
        self.username = username
        self.hashed = hashed

    def checkPassword(self, password):
        return self.hashed == password

class UsernamePassword(CredentialsChecker):

    def __init__(self, username, password):
        self.username = username
        self.password = password

    def checkPassword(self, password):
        return self.password == password

class APOPCredentials(CredentialsChecker):

    def __init__(self, magic, username, digest):
        self.magic = magic
        self.username = username
        self.digest = digest

    def checkPassword(self, password):
        seed = self.magic + password
        m = hashlib.md5()
        m.update(seed)
        myDigest = m.hexdigest()
        return myDigest == self.digest

class MaildirDirdbmDomain(AbstractMaildirDomain):
    """A Maildir Domain where membership is checked by a dirdbm file
    """

    def __init__(self, root=None, postmaster=0):
        """Initialize

        The first argument is where the Domain directory is rooted.
        The second is whether non-existing addresses are simply
        forwarded to postmaster instead of outright bounce

        The directory structure of a MailddirDirdbmDomain is:

        /passwd <-- a dbm file
        /USER/{cur,new,del} <-- each user has these three directories

        A maximum of one 'avatar' (MaildirMailbox) object is maintained
        for each user of this domain.  The avatars are created when the
        first avatarRequest is made.  At present this probably not thread safe.
        """
        AbstractMaildirDomain.__init__(self, root)
        dbm = os.path.join(root, 'passwd')
        if not os.path.exists(root):
            os.makedirs(root)
        self.dbm = anydbm.open(dbm, "c")
        self.postmaster = postmaster
        self.avatarDict = {}

    def userDirectory(self, name):
        """Get the directory for a user

        If the user exists in the dirdbm file, return the directory
        os.path.join(root, name), creating it if necessary.
        Otherwise, returns postmaster's mailbox instead if bounces
        go to postmaster, otherwise return None
        """
        if not self.dbm.has_key(name):
            if not self.postmaster:
                return None
            name = 'postmaster'
        dir = os.path.join(self.root, name)
        if not os.path.exists(dir):
            initializeMaildir(dir)
        return dir
    def close(self):
        self.dbm.close()

    ##
    ## IDomain
    ##
    def addUser(self, user, password):
        self.dbm[user] = password
        # Ensure it is initialized
        self.userDirectory(user)

    def getCredentialsCheckers(self):
        if self._credcheckers is None:
            self._credcheckers = [DirdbmDatabase(self.dbm)]
        return self._credcheckers

    def buildNewMessage(self, user, feeder):
        """Builds a new message into the mailbox for a given user."""
        # Call start message and then loop on feeder until finished calling
        # add line and then eom_message.  Update the list when finished.
        avatar = self.requestAvatar(user)
        if avatar == None:
            return False
        msg = self.startMessage(user)
        while True:
            try:
                msg.lineReceived(feeder.next())
            except:
                break
        result = msg.eomReceived()
        if result:
            avatar.appendMessage(msg)
        return result
        
        
    ##
    ## IRealm
    ##
    def verifyCredentials(self, credentials):
        for checker in self.getCredentialsCheckers():
            try:
                return checker.requestAvatarId(credentials)
            except UnauthorizedLogin:
                continue
        raise UnauthorizedLogin
    
    def requestAvatar(self, avatarId):
        dirname = self.userDirectory(avatarId)
        if dirname == None:
            return None
        if not self.avatarDict.has_key(avatarId):
            self.avatarDict[avatarId] = MaildirMailbox(dirname, avatarId)
        return self.avatarDict[avatarId]

class DirdbmDatabase:

    def __init__(self, dbm):
        self.dirdbm = dbm

    """
    requestAvatarId expects an instance of an object that inherits from
    CredentialsChecker as its paramter.
    """

    def requestAvatarId(self, c):
        if c.username in self.dirdbm:
            if c.checkPassword(self.dirdbm[c.username]):
                return c.username
        raise UnauthorizedLogin()

if __name__ == "__main__":
    # Test code
    class TestFeeder:
        content = ["From: elwynd@example.com",
                   "To: godzilla@mumble.com",
                   "Subject: Test %d",
                   "",
                   "A test message",
                   "Some more text",
                   "Yet more text",
                   "The quick brown fox jumped"]
        def __init__(self):
            self.msg_no = 0

        def get_line(self):
            self.msg_no += 1
            for l in self.content:
                if l[0:7] == "Subject":
                    yield l % (self.msg_no, )
                else:
                    yield l
            
    #testdomain = MaildirDirdbmDomain("/home/elwynd/maildir")
    testdomain = MaildirDirdbmDomain("/tmp")
    testdomain.addUser("postmaster", "postie_pat")
    testdomain.addUser("elwynd", "password")

    magic = "abcdefgh1234567"
    m = hashlib.md5()

    m.update(magic + "password")
    elwynd_digest = m.hexdigest()
    
    cred1 = UsernamePassword("postmaster", "postie_pat")
    cred2 = APOPCredentials(magic, "elwynd", elwynd_digest)
    cred3 = UsernamePassword("postmaster", "not_my_pas")
    cred4 = UsernamePassword("godzilla", "not_my_pas")
    cred5 = APOPCredentials(magic, "elwynd", "0afec56")
    cred6 = APOPCredentials(magic, "mummy", elwynd_digest)

    n = 0
    for i in [cred1, cred2, cred3, cred4, cred5, cred6]:
        n += 1
        try:
            avatarId = testdomain.verifyCredentials(i)
            print "Successful credentials verification: %d %s\n" % (n, avatarId,)
        except:
            print "Verification failed: %d %s\n" %(n, i.username)

    f = TestFeeder()
    g = f.get_line()
    while True:
        try:
            print g.next()
        except:
            break
    # see what is in elwynd's mailbox
    avatar = testdomain.requestAvatar("elwynd")
    orig_list = avatar.listMessages()
    l1 = len(orig_list)
    print "List of messages (%d): " % (l1, ), orig_list

    # Create some more messages for elwynd
    g = f.get_line()
    result = testdomain.buildNewMessage("elwynd", g)
    print "Msg stored: ", result
    g = f.get_line()
    result = testdomain.buildNewMessage("elwynd", g)
    print "Msg stored: ", result
    g = f.get_line()
    result = testdomain.buildNewMessage("elwynd", g)
    print "Msg stored: ", result
    g = f.get_line()
    result = testdomain.buildNewMessage("elwynd", g)
    print "Msg stored: ", result
    new_list = avatar.listMessages()
    l2 = len(new_list)
    print "List of messages (%d): " % (l2, ), new_list
    if l2 == (l1 + 4):
        print "Correct number of messages added"
    else:
        print "New list does not contain expected number of messages"

    # Test getUidl
    print "Uidl: ", avatar.getUidl(3)
    try:
        res = avatar.getUidl(l2+1)
        print "Error: out of range message id did not raise ValueError"
    except ValueError:
        print "Out of range message ifd raised ValueError correctly"

    # Test deletion
    avatar.deleteMessage(4)
    avatar.deleteMessage(3)
    new_list = avatar.listMessages()
    l2 = len(new_list)
    print "List of messages after deletion (%d): " % (l2, ), new_list
    avatar.undeleteMessages()
    new_list = avatar.listMessages()
    l2 = len(new_list)
    print "List of messages after undeletion (%d): " % (l2, ), new_list

    # Test moving new to current
    avatar.moveNewCur(2)
    new_list = avatar.listMessages()
    l2 = len(new_list)
    print "List of messages after moving one message from new to cur (%d): " % (l2, ), new_list
    avatar.moveNewCur()
    new_list = avatar.listMessages()
    l2 = len(new_list)
    print "List of messages after moving all messages from new to cur (%d): " % (l2, ), new_list

    # Test clean trash
    avatar.deleteMessage(2)
    new_list = avatar.listMessages()
    l2 = len(new_list)
    print "List of messages after deleteing one message (%d): " % (l2, ), new_list
    avatar.cleanTrash()
    new_list = avatar.listMessages()
    l2 = len(new_list)
    print "List of messages after cleaning trash (%d): " % (l2, ), new_list

    testdomain.close()
    
    
    
    

    


