#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


"""
DTN-Postfix interface

A single instance of each of the dtn_send_interface dtn_recv_interface classes
are instantiated and run in threads to manage sending and receiving bundles on
the DTN daemon interface. A third thread is used to handle incoming transmission
reports.

The send side handles updating the database that records information about
sent messages.  Bundles are sent as files. Files to be sent are queued by
the Postfix outgoing message handler and the 'retry' timer process.

The report receiver also updates the database as necessary.

The receive side  dumps the bundle payload as a file in the dtn maildir (this is just
a matter of renaming as we get bundles as files).

The send and receive threads use separate connections to the DTN daemon because
of the arcane pseudo-half-duplex nature of the DTN daemon.

At present (until the naming system is improved):
- We expect the local  EID for the gateway to be dtn://gateway.nomadic.n4c.eu
- Oustations will have local  EIDs matching dtn://<user>.nomadic.n4c.eu
- The source EID for this system is <Local EID>/email/out
- Incoming mail bundles should be sent to <Local EID>/email/in
- The reports will be sent back to <Local EID>/email/reports

Permanent registrations will be created for these EIDs with the 'failure action'
set to DEFER so that bundles are retained in case the mail interface is not
active at the time the bundles arrive. We should do this in the dtnd
configuration file.

This piece ofthe program is symmetric between the gateway and the outstation
client implementations except for the destination EID for outgoing bundles.
The local EID name is retrieved from the dtnd to determine the source EID etc.

The sending side receives notications of new emails to send in a Queue.

The receive side informs the email handlers in one of two ways:
- It pushes an 'event' onto a queue (not used in outstations - there may
  not be a POP3 client active when the message is received).
- It keeps a 'state number' which is incremented each time a new email
  is received.  This is prtected by a Lock so that it can be read securely
  by a POP3 server which can then rebuild its message list - except that
  this does not work.  Clients delet mail as they retrieve it, so
  the list contains empty slots and will not be rebuilt easily.  In practice
  we can forget this wholebusiness.  Just wait until the next retrieval cycle.
  I'll leave the code in for the time being in case I think of another way..

Revision History
================
Version   Date       Author         Notes
0.1       14/03/2010 Elwyn Davies   License statement added.
0.0	  25/07/2009 Elwyn Davies   Created for N4C Summer tests 2009
  """

import os
import time
import threading
import dtnapi
import socket
import Queue
import logging
from select import select
from pysqlite2 import dbapi2 as sqlite3

from dp_msg_send_evt import *
import pypop_maildir


# Exception resulting from DTN problems
class DtnError(Exception):
    def __init__(self, reason):
        self.reason = reason
    def __str__(self):
        return "Error communication with DTN daemon: %s" % (repr(self.reason),)

# Local part of source/destination EID
EMAIL_OUT = "email/out"
EMAIL_IN = "email/in"

# Local part of report EID
EMAIL_REPORTS = "email/reports"

# Expiration period for mail bundles
# Try 3 days for now - value in seconds
MAIL_EXPIRY = (3 * 24 * 60 * 60)

# Status report flag values
# See 
REPORT_STATUS_DELIVERED = 8
REPORT_STATUS_DELETED = 16


# Function to make the EIDs (has to be done in each thread)
def build_eids(dtn_handle):
    if dtn_handle == -1:
        raise DtnError("bad DTN handle")
    email_out_addr = dtnapi.dtn_build_local_eid(dtn_handle, EMAIL_OUT)
    email_in_addr = dtnapi.dtn_build_local_eid(dtn_handle, EMAIL_IN)
    report_addr = dtnapi.dtn_build_local_eid(dtn_handle, EMAIL_REPORTS)
    if email_in_addr == None or email_out_addr == None or report_addr == None:
        raise DtnError("failure in dtn_build_local_eid")
    return (email_out_addr, email_in_addr, report_addr)

# Class for dealing with sending email over DTN
# Single instance runs in a thread - only deals with
# sending so is not bound to any address.
class dtn_send(threading.Thread):
    def __init__(self, domain, send_q, logger):
        threading.Thread.__init__(self, name="dtn-send")
        self.domain = domain
        self.send_q = send_q

        # Set up logging functions
        self.logger = logger
        self.loginfo = logger.info
        self.logdebug = logger.debug
        self.logerror = logger.error
        self.logwarn = logger.warn
        
        # Flag to keep the loop running
        self.send_run = True

    def end_run(self):
        self.send_run = False
        
    def run(self):
        # This function loops forever (well more or less) waiting for
        # queued send requests.  So all the variables can be local.
        # Create a connection to the DTN daemon
        dtn_handle = dtnapi.dtn_open()
        if dtn_handle == -1:
            raise DtnError("unable to open connection with daemon")

        # Generate the EIDs
        (email_out_eid, email_in_eid, report_eid) = build_eids(dtn_handle)

        # Open the database connection
        dbconn = None

        # Specify a basic bundle spec
        # - We always send the payload as a permanent file
        pt = dtnapi.DTN_PAYLOAD_FILE
        # - The source address is always the email_addr
        # - The report address is always the report_addr
        # - The registration id needed in dtn_send is the place
        #   where reports come back to.. we always want reports
        #   but it is unclear why we need to subscribe to the 'session'
        #   just to do the send.The reports will come back through
        #   another connection. Lets try with no regid
        regid = dtnapi.DTN_REGID_NONE
        # - The destination address has to be synthesized (later)
        # - We want delivery reports (and maybe deletion reports?)
        dopts = dtnapi.DOPTS_DELIVERY_RCPT
        # - Send with normal priority.
        pri = dtnapi.COS_NORMAL
        # Mail bundlesshould last a while..
        exp = MAIL_EXPIRY

        self.logdebug("Entering DTN send Queue loop")

        # Process mail
        # The thread sits waiting for queued events forever.
        # The thread is terminated by sending a special MSG_END event
        while (self.send_run):
            evt = self.send_q.get()
            if evt.is_last_msg():
                break

            # Build the destination EID
            # The destination may be a list of email addresses if it
            # is going to the gateway but will only be one
            # address if going to another node directly by DTN
            destn = evt.destination()
            self.logdebug("Message for destination(s) %s" % destn)
            if len(destn) == 0:
                self.logerror("No destinations specified: message ignored")
                continue
            if '@' in destn[0]:
                [user, dest_domain] = destn[0].split("@")
            else:
                self.logwarn("Destination %s does not contain a domain name" %
                             destn[0])
                user = destn[0]
                dest_domain = ""
            if dest_domain != self.domain:
                destn_eid = "dtn://gateway.%s/%s" % (self.domain, EMAIL_IN)
            else:
                destn_eid = "dtn://%s.%s/%s" % (user, dest_domain, EMAIL_IN)

            self.loginfo("Sending message to %s" % destn_eid)

            # Send the bundle
            bundle_id = dtnapi.dtn_send(dtn_handle, regid, email_out_eid, destn_eid,
                                        report_eid, pri, dopts, exp, pt,
                                        evt.filename(), "", "")
            if bundle_id == None:
                self.logwarn("Sending of message to %s failed" % destn_eid)
            else:
                # Store the details of the sent bundle
                self.loginfo("%s sent at %d, seq no %d" %(bundle_id.source,
                                                          bundle_id.creation_secs,
                                                          bundle_id.creation_seqno))
        dtnapi.dtn_close(dtn_handle)
        self.loginfo("dtn_send exiting")

# Class for handling incoming email.
# This only deals with incoming email bundles and is bound to
# <local_eid>/email/in
class dtn_receive(threading.Thread):
    def __init__(self, domain, mailbox, recv_q, logger):
        threading.Thread.__init__(self, name="dtn-receive")
        self.domain = domain
        self.mailbox = mailbox
        # recv_q may be None in an outstation
        self.recv_q = recv_q

        # New mail state counter and lock
        self.mail_state = 0
        self.mail_state_lock = threading.Lock()

        # Set up logging functions
        self.logger = logger
        self.loginfo = logger.info
        self.logdebug = logger.debug
        self.logerror = logger.error
        self.logwarn = logger.warn

        # Flag to keep the loop running
        self.receive_run = True

    def end_run(self):
        self.receive_run = False

    def get_mail_state(self):
        self.mail_state_lock.acquire()
        res = self.mail_state
        self.mail_state_lock.release()
        return res
        
    def run(self):
        # This function loops forever (well more or less) waiting for
        # incoming report bundles.
        # This requires it to poll the DTN daemon (yawn).
        # The poll periodically drops out in order to check if the thread has
        # been terminated (this may not be necessary - check if threads hung
        # up in select terminate naturally if the thread is terminated because
        # it is a daemon (but the check is cleaner really).
        
        # Create a connection to the DTN daemon
        dtn_handle = dtnapi.dtn_open()
        if dtn_handle == -1:
            raise DtnError("unable to open connection with daemon")

        # Generate the EIDs
        (email_out_eid, email_in_eid, report_eid) = build_eids(dtn_handle)

        # Open the database connection
        dbconn = None

        # Check if email/in registration exists and register if not
        # Otherwise bind to the existing registration
        regid = dtnapi.dtn_find_registration(dtn_handle, email_in_eid)
        if (regid == -1):
            # Need to register the EID.. make it permanent with 'DEFER'
            # characteristics so that bundles are saved if theye arrive
            # while the handler is inactive
            # Expire the registration a long time in the future
            exp = 365 * 24 * 60 * 60
            # The registration is immediately active
            passive = False
            # We don't want to execute a script
            script = ""
            
            regid = dtnapi.dtn_register(dtn_handle, email_in_eid, dtnapi.DTN_REG_DEFER,
                                        exp, passive, script)
        else:
            dtnapi.dtn_bind(dtn_handle, regid)

        # Wait for 30 seconds before looping
        recv_timeout = 30

        self.loginfo("Entering email receive loop")

        # Now sit and wait for incoming email
        # Note that just using dtn_recv with a timeout doesn't work.
        # The blocking I/O upsets the threading seemingly.
        receive_fd = dtnapi.dtn_poll_fd(dtn_handle)
        while (self.receive_run):
            # Poll currently sets timeout in ms - this is a bug
            dtnapi.dtn_begin_poll(dtn_handle, 5000)
            # The timeout on select is much longer than the dtn_begin_poll
            # timeout, so there is a problem if there is nothing to read
            rd_fd, wr_fd, err_fd = select([receive_fd], [], [], 20)
            if (len(rd_fd) != 1) or (rd_fd[0] != receive_fd):
                # Cancel the poll anyway
                dtnapi.dtn_cancel_poll(dtn_handle)
                raise DtnError("Report select call timed out")
            
            """
            There should always be something to read
            Put in a timeout just in case
            The call to dtn_recv terminates the poll
            We accept the email as a (temporary) file.
            There is a nasty snag here. There is small window
            of opportunity where the bundle has been written
            to a temporary file delivered to this application
            but before file has been renamed and registered in the maildir
            and database. If the application exits during this period
            the email will be lost.  This is not a happy situation.
            At present DTN2 does not have a means for dtn_recv to offer
            a filename that could be used.
            
            Uing DTN_PAYLOAD_MEM would present a similar problem
            During the period the memory image is being written to file
            in the application there is no permanent  copy of the file
            but the DTN daemon believes it has delivered and is busy sending
            a delivery notification.

            NOTE: On receiving the file, the file name is in the bundle
            payload as a NULL terminated string.  Python leaves the terminating
            byte in place.
            """
            email = dtnapi.dtn_recv(dtn_handle, dtnapi.DTN_PAYLOAD_FILE, 1)
            if email != None:
                fn = email.payload
                l = len(fn)
                if fn[l-1] == "\x00":
                    fn = fn[:-1]
                self.logdebug("Got incoming bundle in file %s" % fn)

                # Insert new message file  into the inbox
                final_fn = self.mailbox.newMessageFile(fn)
                if final_fn == None:
                    raise DtnError("Cannot store new mail in inbox")

                self.loginfo("Received email bundle from %s sent %d seq %d" %
                             (email.source,
                              email.creation_secs,
                              email.creation_seqno))
                self.loginfo("New email stored as %s" % final_fn)
                
                # Update database

                # When this is used in a gateway there will be a recv_q
                # where this can be sent off to be fed to Postfix
                # In an outstation, all that needs to be done is flag
                # that a new message has been added to the maildir
                # In practice this may be redundant. Not sure how POP3
                # deals with new mail arriving during a session.
                if self.recv_q != None:                             
                    # Put message on the queue to send it on to Postfix
                    evt = msg_send_evt(MSG_SMTP, final_fn, MSG_NEW, "smtp server")
                    self.recv_q.put_nowait(evt)

                # Increment the state variable - locking while we do
                self.mail_state_lock.acquire()
                self.mail_state += 1
                self.mail_state_lock.release()
                    
            elif dtnapi.dtn_errno(dtn_handle) != dtnapi.DTN_ETIMEOUT:
                raise DtnError("Report: dtn_recv failed with error code %d" %
                               dtnapi.dtn_errno(dtn_handle))
            else:
                pass
                               
        dtnapi.dtn_close(dtn_handle)
        self.loginfo("dtn_receive exiting")
        
# Class for handling incoming reports.
# This only deals with incoming report bundles and is bound to
# <local_eid>/email/reports
class dtn_report(threading.Thread):
    def __init__(self, domain, logger):
        threading.Thread.__init__(self, name="dtn-report")
        self.domain = domain

        # Set up logging functions
        self.logger = logger
        self.loginfo = logger.info
        self.logdebug = logger.debug
        self.logerror = logger.error
        self.logwarn = logger.warn

        # Flag to keep the loop running
        self.report_run = True

    def end_run(self):
        self.report_run = False
        
    def run(self):
        # This function loops forever (well more or less) waiting for
        # incoming report bundles.
        # This requires it to poll the DTN daemon (yawn).
        # The poll periodically drops out in order to check if the thread has
        # been terminated (this may not be necessary - check if threads hung
        # up in select terminate naturally if the thread is terminated because
        # it is a daemon (but the check is cleaner really).
        
        # Create a connection to the DTN daemon
        dtn_handle = dtnapi.dtn_open()
        if dtn_handle == -1:
            raise DtnError("unable to open connection with daemon")

        # Generate the EIDs
        (email_out_eid, email_in_eid, report_eid) = build_eids(dtn_handle)
        
        # Open the database connection
        dbconn = None

        # Check if email/report registration exists and register if not
        # Otherwise bind to the existing registration
        regid = dtnapi.dtn_find_registration(dtn_handle, report_eid)
        if (regid == -1):
            # Need to register the EID.. make it permanent with 'DEFER'
            # characteristics so that bundles are saved if theye arrive
            # while the handler is inactive
            # Expire the registration a long time in the future
            exp = 365 * 24 * 60 * 60
            # The registration is immediately active
            passive = False
            # We don't want to execute a script
            script = ""
            
            regid = dtnapi.dtn_register(dtn_handle, report_eid, dtnapi.DTN_REG_DEFER,
                                        exp, passive, script)
        else:
            dtnapi.dtn_bind(dtn_handle, regid)

        # Wait for 30 seconds before looping
        recv_timeout = 30

        self.loginfo("Entering DTN report recv loop")

        # Now sit and wait for incoming reports
        # Note that just using dtn_recv with a timeout doesn't work.
        # The blocking I/O upsets the threading seemingly.
        report_fd = dtnapi.dtn_poll_fd(dtn_handle)
        while (self.report_run):
            # Poll currently sets timeout in ms - this is a bug
            dtnapi.dtn_begin_poll(dtn_handle, 5000)
            # The timeout on select is much longer than the dtn_begin_poll
            # timeout, so there is a problem if there is nothing to read
            rd_fd, wr_fd, err_fd = select([report_fd], [], [], 20)
            if (len(rd_fd) != 1) or (rd_fd[0] != report_fd):
                # Cancel the poll anyway
                dtnapi.dtn_cancel_poll(dtn_handle)
                raise DtnError("Report select call timed out")
            
            # There should always be something to read
            # Put in a timeout just in case
            # The call to dtn_recv terminates the poll
            report = dtnapi.dtn_recv(dtn_handle, dtnapi.DTN_PAYLOAD_MEM, 1)
            if report != None:
                self.logdebug("Received bundle")
                if report.status_report != None:
                    self.logdebug("Received status report")
                    if report.status_report.flags == REPORT_STATUS_DELIVERED:
                        self.loginfo( "Received delivery report re from %s sent %d seq %d" %
                                      (report.status_report.bundle_id.source,
                                       report.status_report.bundle_id.creation_secs,
                                       report.status_report.bundle_id.creation_seqno))
                        # Update the database
      
                    elif report.status_report.flags == REPORT_STATUS_DELETED:
                        self.loginfo("Received deletion report re from %s sent %d seq %d" %
                                     (report.status_report.bundle_id.source,
                                      report.status_report.bundle_id.creation_secs,
                                      report.status_report.bundle_id.creation_seqno))
                        # Update the database

                    else:
                        self.logwarn("Received unexpected report: Flags: %d" % report.status_report.flags)

            elif dtnapi.dtn_errno(dtn_handle) != dtnapi.DTN_ETIMEOUT:
                self.logerror("dtn_recv failed with error code %d" %
                              dtnapi.dtn_errno(dtn_handle))
                raise DtnError("Report: dtn_recv failed with error code %d" %
                               dtnapi.dtn_errno(dtn_handle))
            else:
                pass
                               
        dtnapi.dtn_close(dtn_handle)
        self.loginfo("dtn_report exiting")
            
# Test code
if __name__ == "__main__":
    logger = logging.getLogger("test")
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    fmt = logging.Formatter("%(levelname)s %(threadName)s %(message)s")
    ch.setFormatter(fmt)
    logger.addHandler(ch)
    
    dtn_send_q = Queue.Queue()
    dtn_recv_q = Queue.Queue()
    nomadic_domain = "nomadic.n4c.eu"
    tmp_domain = pypop_maildir.MaildirDirdbmDomain("/tmp")
    if not tmp_domain.userDirectory("dtnin"):
        tmp_domain.addUser("dtnin", "password")
    dtnin_mailbox = pypop_maildir.MaildirMailbox(tmp_domain.userDirectory("dtnin"),
                                                 "dtnin")
    
    # Sending thread
    dtn_send_handler = dtn_send(nomadic_domain, dtn_send_q, logger)
    dtn_send_handler.setDaemon(True)
    dtn_send_handler.start()

    # Receiving thread
    dtn_recv_handler = dtn_receive(nomadic_domain, dtnin_mailbox,
                                   dtn_recv_q, logger)
    dtn_recv_handler.setDaemon(True)
    dtn_recv_handler.start()

    # Report handler thread
    dtn_report_handler = dtn_report(nomadic_domain, logger)
    dtn_report_handler.setDaemon(True)
    dtn_report_handler.start()

    time.sleep(0.1)
    logger.info("Send handler running: %s" % dtn_send_handler.getName())
    logger.info("Receive handler running: %s" % dtn_recv_handler.getName())
    logger.info("Report handler running: %s" % dtn_report_handler.getName())

    evt = msg_send_evt(MSG_DTN,
                       "/etc/group",
                       MSG_NEW,
                       ["elwynd@nomadic.n4c.eu"])
    dtn_send_q.put_nowait(evt)

    logger.info("Waiting for incoming mail")
    evt = dtn_recv_q.get()
    logger.info("Mail received: ", evt.filename())

    try:
        while True:
            pass
    except:
        evt = msg_send_evt(MSG_END,
                           "",
                           MSG_NEW,
                           "")
        dtn_send_q.put_nowait(evt)
        dtn_recv_handler.end_run()
        dtn_report_handler.end_run()
        time.sleep(1)
        

        
    

                                 
        

         

        
