#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

"""
This is the controller for the DTN nomadic email system that is loaded into
'outstations' i.e. 'small' systems where there is generallyno connection to
the Internet and there is no way of  making an SMTP connection to the rest of the world
for most of the time.

It links between an email clinet (user agent) such as Thunderbird or Evolution
and the DTN daemon.

It provides
- a POP server for incoming email
- an SMTP server for outgoing email
- interfaces to convert between emails and DTN bundles

It is intended at present to work for a single DTN domain or Communications
Association. An outstation deals with only a single email address at present
due to limitations of the DTN naming system in DTN2 (as of version 2.6.0+), and
the email address is closely mapped to the EID of the node where the outstation
is installed: user@nomadic.n4c.eu <-> dtn://user.nomadic.n4c.eu.

It stores emails in two maildir structures - one for incoming and one for 
outgoing.Outgoing emails are maintained until either a delivery report is sent
back or a delivery timeout occurs.  If the bundle sent for an email willhave expired
without a delivery report being returned, the bundle is retransmitted
[TBD Does custody do this for us?  Yes.. but only if every node will take custody.]

We also need to cope with possible incoming duplicates as the same bundle may be
delivered by multiple paths.

The application is heavily multi-threaded:
- Main thread.. this controller
- One master SMTP server listener thread plus a thread for each outgoing email connection.
- One master POP3 server listener thread plus a thread for each POP3 connection
- Three DTN handler threads dealing with:
  - Sending outgoing emails as bundles
  - Receiving incoming emails as bundles
  - Receiving delivery reports.
- Retransmit and junk disposal management thread

Communication between threads in the outstation (different from the gateway)
uses a Queue to notify the DTN sender thread of new mails to send out as bundles
(note that in general the bundle daemon will accept custody of the bundle and
send it as and when there is a connection - give or take lack of storage, the
DTN sender will not need to retry sending.  On the incoming side, new messages
are dumped into the dtnin (or user name) maildir mailbox.  This process is
thread safe as the files all get unique names.  The POP3 server thread is
informed that there has been a change and will remake its list next time a
command is processed.  All that is needed is an incrementing counter.  We won't
worry about wrap around at the moment - even the most enthusiastic emailer and
spammers are unlikely to cause a problem  here. The counter is only incremented
by the DTN receive thread and read by the POP3 threads.  It is protected by a
simple Lock. The critical sections will be very short so no need to worry about
deadlock or threads being held off for a long time. The Queue used in the gateway
on the DTN receiving side is not used here - all the communication is effectively
through instances of the maildir mailbox. The counter and the lock are hidden
in the DTN receiver class.

NOTE:
The SMTP and POP3 server threads make use of the Python SocketServer framework.
Frankly, the code in the Python 2.5 library is complete rubbish.  It is
esentially impossible to set the reuse address flag making (at least) testing
very difficult.  Further it is extremely difficult to shutdown the servers
cleanly.  Most of this is fixed in the Python 2.6 release. Since this module
is pure Python and doesn't do anything v2.6 specific itself, I have extracted
the module and am using for this program.

Revision History
================
Version   Date       Author         Notes
0.1       14/03/2010 Elwyn Davies   License statement added.
0.0	  25/07/2009 Elwyn Davies   Created for N4C Summer tests 2009
"""
import os
import time
import threading
import Queue
import socket
import SocketServer
import logging
import logging.handlers
import logging.config

import pypop_maildir
import dp_pop3
import dp_smtpserv
import dp_dtn
from dp_msg_send_evt import *
import dtnapi

def main(mailbox_root, log_file, log_config_file):
    # Setup logging
    logging.config.fileConfig(log_config_file)
    pymail_logger = logging.getLogger("dtn_pymail")
    loginfo = pymail_logger.info
    logdebug = pymail_logger.debug

    # Set domain for which we are handling mail
    mail_domain = "nomadic.n4c.eu"
    
    # Check mailboxes dtnin and dtnout  exist and create if not ***TBD***
    nomadic_domain = pypop_maildir.MaildirDirdbmDomain(mailbox_root)
    if not nomadic_domain.userDirectory("dtnin"):
        nomadic_domain.addUser("dtnin", "password")
    if not nomadic_domain.userDirectory("dtnout"):
        nomadic_domain.addUser("dtnout", "password")
    dtnin_mailbox  = pypop_maildir.MaildirMailbox(nomadic_domain.userDirectory("dtnin"),
                                                  "dtnin")
    dtnout_mailbox = pypop_maildir.MaildirMailbox(nomadic_domain.userDirectory("dtnout"),
                                                  "dtnout")

    # Create queue to send requests for messages to be written out by DTN
    # FIFO queue with no depth limitation at present
    dtn_send_q = Queue.Queue()

    #====================================================================#
    # Create threads to handle DTN interface
    # There are three threads that desl with:
    # - sending emails
    # - receiving emails
    # - receiving reports

    # Sending thread
    dtn_send_handler = dp_dtn.dtn_send(mail_domain, dtn_send_q, pymail_logger)
    dtn_send_handler.setDaemon(True)
    dtn_send_handler.start()

     # Receiving thread
    dtn_receive_handler = dp_dtn.dtn_receive(mail_domain, dtnin_mailbox,
                                             None, pymail_logger)
    dtn_receive_handler.setDaemon(True)
    dtn_receive_handler.start()
    
    # Report handler thread
    dtn_report_handler = dp_dtn.dtn_report(mail_domain, pymail_logger)
    dtn_report_handler.setDaemon(True)
    dtn_report_handler.start()

    #====================================================================#
    # Create listener for server(s) to handle outgoing mails via SMTP
    smtp_listener = dp_smtpserv.dtn_smtp_server(mail_domain, dtnout_mailbox,
                                                dtn_send_q, pymail_logger)

    # Start a thread with the listener -- that thread will then start one
    # more thread for each SMTP connection request
    # This is usually one per email, but not necessarily.
    smtp_listener_thread = threading.Thread(target=smtp_listener.serve_forever,
                                            name="SMTP Listener")

    # Exit the listener thread when the main thread terminates
    smtp_listener_thread.setDaemon(True)
    smtp_listener_thread.start()

    #====================================================================#
    # Create server to provide POP3 service for mailbox on the incoming
    # side.
    pop3_listener = dp_pop3.dtn_pop3_server(nomadic_domain,
                                            "dtnin",
                                            pymail_logger)

    # Start a thread with the listener -- that thread will then start one
    # more thread for each POP3 connection request
    pop3_listener_thread = threading.Thread(target=pop3_listener.serve_forever,
                                            name="POP3 Listener")

    # Exit the listener thread when the main thread terminates
    pop3_listener_thread.setDaemon(True)
    pop3_listener_thread.start()

    
    #====================================================================#

    # Let everything have a chance to get going (old trick)
    time.sleep(0.1)
    loginfo("DTN Send handler running in thread: %s" % dtn_send_handler.getName())
    loginfo("DTN Receive handler running in thread: %s" % dtn_receive_handler.getName())
    loginfo("DTN Report handler running in thread: %s" % dtn_report_handler.getName())   
    loginfo("SMTP Listener loop running in thread:%s" % smtp_listener_thread.getName())
    loginfo("POP3 Listener loop running in thread:%s" % pop3_listener_thread.getName())

    try:
        while True:
            pass
    except:
        # Shutdown
        # Think a bit about ordering here...
        # Shutdown the producer threads first
        dtn_receive_handler.end_run()
        smtp_listener.end_run()

        # Give them a few seconds to die...
        time.sleep(6)

        # Then shutdown the consumer threads
        pop3_listener.end_run()
        dtn_report_handler.end_run()
        evt = msg_send_evt(MSG_END,
                           "",
                           MSG_NEW,
                           "")
        dtn_send_q.put_nowait(evt)

        # ..and wait for them to die...
        time.sleep(2)                                    
        

if __name__ == "__main__":
    main("/home/user/dtn/maildir", "/media/mmc2/dtn/log", "/home/user/dtn/python/dtn_pymail_log.conf")
    # Test code


