#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


# Simple POP3 server

"""
Post-office Protocol version 3 Server (Threading version) 

@author: Elwyn Davies

This is a pretty complete but straightforward implementation of RFC 1939 plus
a few of the options from RFC 2449.

It implements USER/PASS and APOP 'authentication'.
It implements the basic commands (STAT, LIST, RETR, DELE, NOOP and RSET) plus
the a subset of the optional commands (TOP, UIDL, CAPA, IMPLEMENTATION).
It does not provide SASL, PIPELINING, RESP-CODES, LOGIN-DELAY or EXPIRE.

In particular the original purpose for this code was to provide an interface
to a local mailbox on a Symbian mobile phone.  The connection would be internal
to the phone which makes the use of SSL and complex authentication superfluous.
I has now moved on in this version to supply a POP server on a general purpose
Linux box connected in a multithreaded environment where other threads will provide
DTN connectivity and an SMTP server.

It is intended to connect to a maildir mailbox implemented as local files on the
phone so that additional security beyond the access controls on the overall
phone make little sense.

Because of the usual way that POP3 is used (open a session, see what emails
are listed, retrieve them and delete them in turn and then quit the session)
there is little point (and some difficulty) in trying to inform the POP3
server if new files are added to the maildir during a session.  They are just
picked up on the next session.

It was written to support implementation of the DTN bundle protocol on the
mobile phone, which would deliver mail messages in bundles and store them in the
maildir.  The POP3 server would allow the files to be viewed through the standard
email client on the phone just like any other mailbox, but it would be local to
the phone (accessed using TCP on localhost). The limitations of the mobile phone
email client make pipelining irrelevant, and because normally there will be only
one user who might have more then one mailbox active but can only single task on
them, LOGIN-DELAY is not useful. 

The code can support access to multiple mailboxes in parallel if required.  In
principle the mailbox can be accessed from clients off the phone, but it is probably
sensible to limit accees to localhost (127.0.0.1) only.

The overall structure is based on the SocketServer module in Python with a threading
Listener. There is a listener (normally on port 110) that spawns a server object
in a spearate thread when a connections is made.  There is a single maildomain
with potentially multiple mailboxes owned by different names.  The mailbox is
connected up to the server when there has been an authenticated login.

TO-DO:
Login timeout.
Restrict login to localhost

Acknowledgement:
The structure of this software was inspired by the Twisted Python code for the
POP3 server and maildir structure.  It has, however, been hacked so that there
is very little of the original Twisted code left on account of not running with
the Twisted event framework and requiring a leaner solution. 

Revision History
================
Version   Date       Author         Notes
0.1       14/03/2010 Elwyn Davies   License statement added.
0.0	  25/07/2009 Elwyn Davies   Created for N4C Summer tests 2009
"""

import string
#import warnings
import socket
import hashlib
import time
import os
import random
from SocketServer2_6 import *
import threading
import logging
import sys
from  select import select

on_symbian = (sys.platform == "symbian_s60")

if on_symbian:
    import globalui

import pypop_maildir


class _POP3MessageDeleted(Exception):
    """
    Internal control-flow exception.  Indicates the file of a deleted message
    was requested.
    """

class POP3Error(Exception):
    pass

class _IteratorBuffer(object):
    bufSize = 0

    def __init__(self, write, iterable, memoryBufferSize=None):
        """
        Create a _IteratorBuffer.

        @param write: A one-argument callable which will be invoked with a list
        of strings which have been buffered.

        @param iterable: The source of input strings as any iterable.

        @param memoryBufferSize: The upper limit on buffered string length,
        beyond which the buffer will be flushed to the writer.
        """
        self.lines = []
        self.write = write
        self.iterator = iter(iterable)
        if memoryBufferSize is None:
            memoryBufferSize = 2 ** 16
        self.memoryBufferSize = memoryBufferSize


    def __iter__(self):
        return self


    def next(self):
        try:
            v = self.iterator.next()
        except StopIteration:
            if self.lines:
                self.write(self.lines)
            # Drop some references, in case they're edges in a cycle.
            del self.iterator, self.lines, self.write
            raise
        else:
            if v is not None:
                self.lines.append(v)
                self.bufSize += len(v)
                if self.bufSize > self.memoryBufferSize:
                    self.write(self.lines)
                    self.lines = []
                    self.bufSize = 0


# Cache the hostname (XXX Yes - this is broken)
if False: #platform.isMacOSX():
    # On OS X, getfqdn() is ridiculously slow - use the
    # probably-identical-but-sometimes-not gethostname() there.
    DNSNAME = socket.gethostname()
else:
    DNSNAME = socket.getfqdn()

def idGenerator():
    i = 0
    while True:
        yield i
        i += 1

def messageid(uniq=None, N=idGenerator().next):
    """Return a globally unique random string in RFC 2822 Message-ID format

    <datetime.pid.random@host.dom.ain>

    Optional uniq string will be added to strenghten uniqueness if given.
    """
    datetime = time.strftime('%Y%m%d%H%M%S', time.gmtime())
    pid = os.getpid()
    rand = random.randrange(2**31L-1)
    if uniq is None:
        uniq = ''
    else:
        uniq = '.' + uniq

    return '<%s.%s.%s%s.%s@%s>' % (datetime, pid, rand, uniq, N(), DNSNAME)



class FileIterator:
    """
    Takes an open file object and provides a generator which iterates over
    the lines terminating them with "\r\n" plus the terminator line when
    EOF is reached. Move the file out of 'new' into 'cur' after sending.
    """
    def __init__(self, mbox, msg_no, fp):
        self.fp = fp
        self.mbox = mbox
        self.msg_no = msg_no
        self.eof_seen = False

    def feedLine(self):
        if not self.eof_seen:
            while True:
                buffer = self.fp.readline(1024)
                if buffer != "":
                    if buffer[0] == ".":
                        byte_stuff = "."
                    else:
                        byte_stuff = ""
                    if buffer.endswith("\r\n"):
                        yield byte_stuff + buffer
                    elif buffer.endswith("\n"):
                        yield byte_stuff + buffer[:-1] + "\r\n"
                    else:
                        yield byte_stuff + buffer + "\r\n"
                else:
                    self.eof_seen = True
                    yield ".\r\n"
                    break
        else:
            raise StopIteration

    def closeFile(self):
        self.fp.close()
        self.mbox.moveNewCur(self.msg_no)


class Pop3Handler(BaseRequestHandler):
    """
    POP3 server protocol implementation.

    @cvar delimiter: The line-ending delimiter to use. By default this is
                     '\\r\\n'.
    @cvar MAX_LENGTH: The maximum length of a line to allow (If a
                      sent line is longer than this, the connection is dropped).
                      Default is 16384.

    @ivar timeOut: An integer which defines the minimum amount of time which
    may elapse without receiving any traffic after which the client will be
    disconnected.

    """

    delimiter = '\r\n'
    listEnder = '.'
    MAX_LENGTH = 16384

    AUTH_CMDS = ['CAPA', 'USER', 'PASS', 'APOP', 'AUTH', 'RPOP', 'QUIT']
    TRANS_CMDS = ['CAPA', 'STAT', 'LIST', 'RETR', 'DELE', 'NOOP', 'RSET',
                  'TOP', 'LAST', 'UIDL', 'QUIT', 'STOP']

    def setup(self):
        self.connection = self.request
        self.wfile = self.connection.makefile('wb', 0)

    def finish(self):
        if not self.wfile.closed:
            self.wfile.flush()
        self.wfile.close()

    def handle(self):
        self.pop3_thread = threading.currentThread()
        self.pop3_thread.setName("POP3 server - %d" %
                                 self.server.next_server_num)
        self.server.next_server_num += 1

        # Tell listener we are running
        self.server.add_thread(self)
        
        # Logging functions
        self.loginfo = self.server.logger.info
        self.logdebug = self.server.logger.debug
        self.logwarn = self.server.logger.warn
        self.logerror = self.server.logger.error

        self.loginfo("New POP3 connection from %s" % self.client_address[0])

        # Magic number used as nonce for APOP hashing
        self.magic = self.generateMagic()
        self.successResponse(self.magic)

        # The Domain
        self.maildomain = self.server.domain

        # The current user
        self.userIs = None

        #Clear up routine
        self._onLogout = None

        # Set this pretty low -- POP3 clients are expected to log in, download
        # everything, and log out.
        self.timeOut = 300

        # Current protocol state
        self.state = "AUTH"

        # Message index of the highest retrieved message.
        self._highest = 0

        # Terminates loop when there is no to come on there has been an error
        self.pop3_run = True

        # Mailbox used
        self.mbox = None
        self.mail_state = 0

        # Reading buffer
        buf = ""

        # File descriptor for basuc  socket
        receive_fd = self.request.fileno()

        # Loop reading lines until EOF
        while self.pop3_run:
            # Check if we have a line buffered up (This is very unlikely in POP3)
            i = buf.find('\n')
            if i == -1:
                # Read some more
                # Drop out periodically to check we haven't been stopped
                rd_fd, wr_fd, err_fd = select([receive_fd], [], [receive_fd], 5)
                if (len(err_fd) > 0) and (err_fd[0] == receive_fd):
                    self.logdebug("Socket closed")
                    self.pop3_run = False
                    continue
                if (len(rd_fd) != 1) or (rd_fd[0] != receive_fd):
                    if len(rd_fd) != 0:
                        self.logerror("Spurious output from select")
                    # Loop round again
                    continue
                more = self.request.recv(1024)
                if more == "":
                    self.logdebug("EOF occurred")
                    self.pop3_run = False
                    continue
                buf += more
                i = buf.find('\n')
                if i == -1:
                    self.logdebug("Stuck with partial line")
                    continue
            if i == len(buf) - 1:
                line = buf
                buf = ""
            else:
                line = buf[:i+1]
                buf = buf[i+1:]
                        
            # Analyse incoming message
            line = line.rstrip(self.delimiter)
            self.logdebug("In state %s Line received: %s" % (self.state, line))
            if len(line) >self.MAX_LENGTH:
                self.connectionLost("Line length exceeded")
                break

            # Process line
            self.lineReceived(line)

        # Finished
        self.server.remove_thread(self)
        self.finish()

    def connectionLost(self, reason):
        if self._onLogout is not None:
            self._onLogout()
            self._onLogout = None
        # self.setTimeout(None)
        self.loginfo("Ending connection: %s" % reason)
        self.pop3_run = False

    def sendLine(self, line):
        """Sends a line to the other end of the connection.
        """
        self.logdebug("POP3 S: %s" % repr(line))
        if line[0] == self.listEnder[0]:
            byteStuff = self.listEnder
        else:
            byteStuff = ""
        return self.wfile.write(byteStuff + line + self.delimiter)

    def sendTerminator(self):
        self.logdebug("POP3 S: .")
        return self.wfile.write(self.listEnder + self.delimiter)

    def lineReceived(self, line):
        self.logdebug("POP3 C: %s" % repr(line))
        # self.resetTimeout()
        getattr(self, 'state_' + self.state)(line)

    def generateMagic(self):
        return messageid()

    def successResponse(self, message=''):
        self.sendLine("+OK " + str(message))

    def failResponse(self, message=''):
        self.sendLine('-ERR ' + str(message))

    def state_AUTH(self, line):
        try:
            return self.processCommand(self.AUTH_CMDS, *line.split(' '))
        except (ValueError, AttributeError, POP3Error, TypeError), e:
            self.logerror("Bad protocol or server: %s %s" %
                          (e.__class__.__name__, e))
            self.failResponse('bad protocol or server: %s: %s' %
                              (e.__class__.__name__, e))
            return None

    def state_TRANSACTION(self, line):
        try:
            return self.processCommand(self.TRANS_CMDS, *line.split(' '))
        except (ValueError, AttributeError, POP3Error, TypeError), e:
            if isinstance(e, POP3Error):
                self.failResponse(e)
            else:
                self.logerror("Bad protocol or server: %s %s" %
                              (e.__class__.__name__, e))
                self.failResponse('bad protocol or server: %s: %s' %
                                  (e.__class__.__name__, e))
                return None

    def processCommand(self, valid_cmds, command, *args):
        command = string.upper(command)
        validCmd = command in valid_cmds
        if not validCmd:
            if self.mbox:
                raise POP3Error("cannot do " + command + " while logged in")
            else:
                raise POP3Error("not authenticated yet: cannot do " + command)
        self.logdebug("Processing command %s" % command)
        f = getattr(self, 'do_' + command, None)
        if f:
            return f(*args)
        raise POP3Error("Unknown protocol command: " + command)


    def listCapabilities(self):
        baseCaps = [
            "TOP",
            "USER",
            "UIDL",
            "LAST",
            "IMPLEMENTATION Folly Consulting Python POP3 v0.0"
        ]

        return baseCaps

    def do_CAPA(self):
        self.successResponse("I can do the following:")
        for cap in self.listCapabilities():
            self.sendLine(cap)
        self.sendTerminator()

    def do_AUTH(self, args=None):
        self.failResponse("AUTH extension unsupported")
        return

    def do_APOP(self, user, digest):

        self.cred = pypop_maildir.APOPCredentials(self.magic, user, digest)
        try:
            # Get the mailbox associated with this user
            # The mailboxes are cached per user
            avatarId = self.maildomain.verifyCredentials(self.cred)
            self.mbox = self.maildomain.requestAvatar(avatarId)
            # Update the message list in case anything new has arrived
            #This has to be done only after a new login
            # Otherwise POP gets confused.
            self.mbox.makeMessageList()
            self._onLogout = self.logout
            self.state = "TRANSACTION"
            self.successResponse('Authentication succeeded')
            self.loginfo("Authenticated login for " + user)
        except pypop_maildir.UnauthorizedLogin:
            self.logwarn("Denied login attempt from " + user)
            self.failResponse("Login failed")

    def do_USER(self, user):
        # Only dealing with user matched to DTN EID
        if user != self.server.user:
            self.failResponse("User not acceptable")
            return

        self.userIs = user
        self.successResponse('USER accepted, send PASS')

    def do_PASS(self, password):
        if self.userIs is None:
            self.failResponse("USER required before PASS")
            return
        user = self.userIs
        self.userIs = None
        self.cred = pypop_maildir.UsernamePassword(user, password)
        try:
            # Get the mailbox associated with this user
            # The mailboxes are cached per user
            avatarId = self.maildomain.verifyCredentials(self.cred)
            self.mbox = self.maildomain.requestAvatar(avatarId)
            # Update the message list in case anything new has arrived
            # This has to be done only after a new login
            # Otherwise POP gets confused.
            self.mbox.makeMessageList()
            self._onLogout = self.logout
            self.state = "TRANSACTION"
            self.successResponse('Authentication succeeded')
            self.loginfo("Authenticated login for " + user)
        except pypop_maildir.UnauthorizedLogin:
            self.logwarn("Denied login attempt from " + user)
            self.failResponse("Login failed")


    def do_STAT(self):
        msg_list = self.mbox.listMessages()
        n = 0
        s = 0
        for m in msg_list:
            if m > 0:
                n += 1
                s += m
        self.successResponse("%d %d" %(n, s))

    def do_LIST(self, i=None):
        if i is None:
            msg_list = self.mbox.listMessages()
            self.successResponse("%d messages" % len(msg_list))
            n = 0
            s = 0
            for m in msg_list:
                n += 1
                if m > 0:
                    self.sendLine("%d %d" %(n, m))
            self.sendTerminator()
        else:
            try:
                i = int(i)
                if i < 1:
                    raise ValueError()
                else:
                    sz = self.mbox.listMessages(i-1)
                    if sz == 0:
                        raise ValueError
                    else:
                        self.successResponse("%d %d" % (i, sz))
            except ValueError:
                self.failResponse("Invalid message-number: %r" % (i,))

    def do_UIDL(self, i=None):
        if i is None:
            self.successResponse("unique-id listing follows")
            msg_list = self.mbox.listMessages()
            n = 0
            s = 0
            for m in msg_list:
                n += 1
                uidl = self.mbox.getUidl(n - 1)
                self.sendLine("%d %s" % (n, uidl))
            self.sendTerminator()
        else:
            try:
                i = int(i)
                uidl = self.mbox.getUidl(i-1)
                self.successResponse("%d %s" % (i, uidl))
            except ValueError:
                self.failResponse("Invalid message-number: %d" % (i,))

    def do_TOP(self, i=None, size=None):
        if i == None or size == None:
            self.failResponse("message-id and line count are mandatory for TOP")
            return
        try:
            size = int(size)
        except:
            self.failResponse("TOP - line count must be an integer")
        else:
            if size < 0:
                self.failResponse("TOP - line count must be positive")
                return
            try:
                i = int(i)
                f = self.mbox.getMessage(i-1)
            except:
                self.failResponse("TOP - bad message number")
            else:
                self.successResponse("Top %d lines of message %d follow" % (size, i))
                done = False
                amount = 1024
                headers = True
                linecount = 0
                buf = ''
                while not done:
                    data = f.read(amount)
                    if data == "":
                        done = True
                        break
                    buf += data
                    if headers:
                        # Look for the blank line between headers and body
                        df, sz = buf.find('\r\n\r\n'), 4
                        if df == -1:
                            df, sz = buf.find('\n\n'), 2
                        if df != -1:
                            linecount = 1
                            headers = False
                        else:
                            df, sz = buf.rfind('\n'), 1
                        if df != -1:
                            # Send any complete header lines
                            for s in buf[:df].splitlines():
                                self.sendLine(s)
                            df += sz
                            buf = buf[df:]
                    if linecount > 0:
                        self.logdebug("%s" % buf)
                        dsplit = buf.splitlines()
                        if len(dsplit) > 0:
                            self.buf = dsplit[-1]
                        else:
                            self.buf = ""
                        for ln in dsplit[:-1]:
                            if linecount > size:
                                self.done = True
                                break
                            self.sendLine(ln)
                            linecount += 1
                f.close()
                self.sendTerminator()

    def do_RETR(self, i=None):
        if i == None:
            self.failResponse("message number mandatory for RETR")
        else:
            try:
                i = int(i)
                f = self.mbox.getMessage(i-1)
                self.successResponse("Message follows")
                self._highest = max(self._highest, i)
                eof_seen = False
                while not eof_seen:
                    buffer = f.readline(1024)
                    if buffer != "":
                        if buffer[0] == ".":
                            byte_stuff = "."
                        else:
                            byte_stuff = ""
                        if buffer.endswith("\r\n"):
                            self.wfile.write(byte_stuff + buffer)
                        elif buffer.endswith("\n"):
                            self.wfile.write(byte_stuff + buffer[:-1] + "\r\n")
                        else:
                            self.wfile.write(byte_stuff + buffer + "\r\n")
                    else:
                        eof_seen = True
                        self.wfile.write(".\r\n")
                        f.close()
                        self.mbox.moveNewCur(i-1)
            except:
                self.failResponse("Bad message number")

    def do_DELE(self, i=None):
        if i == None:
            self.failResponse("message-id mandatory for DELE")
            return
        try:
            i = int(i)-1
            self.mbox.deleteMessage(i)
            self.successResponse()
        except:
            self.failResponse("Delete failed")


    def do_NOOP(self):
        """Perform no operation.  Return a success code"""
        self.successResponse("NOOP did not do anything")


    def do_RSET(self):
        """Unset all deleted message flags"""
        try:
            self.mbox.undeleteMessages()
        except:
            log.err()
            self.failResponse("unable to restore messages")
        else:
            self._highest = 0
            self.successResponse("messages restored")


    def do_LAST(self):
        """
        Return the index of the highest message yet downloaded.
        """
        self.successResponse(self._highest)


    def do_RPOP(self, user):
        self.failResponse('permission denied, sucker')

    def do_QUIT(self):
        if self.mbox:
            try:
                self.mbox.sync()
                self.mbox.moveNewCur()
                self.mbox.cleanTrash()
                self.successResponse("mailbox tidied on QUIT")
                self.mbox=None
                self._onLogout = None
            except:
                self.failResponse("problem encountered during QUIT")
            finally:
                self.mbox = None
        else:
            self.successResponse("connection dropped when not logged in")
        self.connectionLost("Finished")

    def do_STOP(self):
        # Non-standard command.  This causes the whole server to shutdown.
        # It is needed as a way of terminating the server in the mobile phone
        # environment.  We use the coarse debugging function, although it will
        # be a reasonably graceful shutdown.
        if on_symbian:
            globalui.global_note(u"POP3 server terminating")
        else:
            self.loginfo("POP3 server terminating")        

    def logout(self):
        if self.mbox:
            self.mbox.sync()

class Pop3Listener(ThreadingMixIn, TCPServer):

    def __init__(self, addr, handler, 
                 domain, user, logger):        
        # These are used  by individual requests
        # accessed via self.server in the handle function
        self.domain = domain
        self.user = user
        self.logger = logger
        self.running_threads = set()
        self.next_server_num = 1

        # Setup to produce a daemon thread for each incoming request
        TCPServer.__init__(self, addr, handler,
                           bind_and_activate=False)
        self.allow_reuse_address = True
        self.server_bind()
        self.server_activate()
                         
        self.daemon_threads = True

    def add_thread(self, thread):
        self.logger.debug("New thread added")
        self.running_threads.add(thread)

    def remove_thread(self, thread):
        if thread in self.running_threads:
            self.running_threads.remove(thread)

    def end_run(self):
        for thread in self.running_threads:
            if thread.pop3_thread.isAlive():
                thread.connectionLost("Listener shutting down")
        del self.running_threads
        self.shutdown()

def dtn_pop3_server(domain, user, logger):
    HOST, PORT = "localhost", 2110
    return Pop3Listener((HOST, PORT), Pop3Handler,
                        domain, user, logger)
# Test code
if __name__ == "__main__":
    import sys
    
    logger = logging.getLogger("test")
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    fmt = logging.Formatter("%(levelname)s %(threadName)s %(message)s")
    ch.setFormatter(fmt)
    logger.addHandler(ch)
    
    pop3Domain = pypop_maildir.MaildirDirdbmDomain("/tmp",
                                                   postmaster=0)
    if not pop3Domain.userDirectory("dtnin"):
        pop3Domain.addUser("dtnin", "password")
        
    server = Pop3Listener(("localhost", 2110), Pop3Handler,
                          pop3Domain, "dtnin", logger)

    server_thread = threading.Thread(target=server.serve_forever)
    server_thread.setDaemon(True)
    server_thread.start()
    logger.info("Server running in thread: %s" %server_thread.getName())

    # Change this to True to just create a standalone POP3 server for testing
    if False:
        try:
            while True:
                pass
        except:
            pass

        server.end_run()

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect(("localhost", 2100))
    rfile = sock.makefile("r")
    wfile = sock.makefile("w")

    def r(ss):
        logger.info("Send %s" % ss)
        wfile.write(ss+"\r\n")
        wfile.flush()
        response = rfile.readline(1024).rstrip("\r\n")
        logger.debug("Receive: %s" % response)
            
    def rm(ss):
        logger.info("Send %s" % ss)
        wfile.write(ss+"\r\n")
        wfile.flush()
        response = rfile.readline(1024).rstrip("\r\n")
        logger.debug("Receive: %s" % response)
        if response[:4]  == "+OK ":
            logger.debug("Receiving multi-line response:")
            while True:
                response = rfile.readline(1024).rstrip("\r\n")
                logger.debug("    %s" % response)
                if response == ".":
                    break
        else:
            logger.debug("Received: %s", resp)
            
    
    resp = rfile.readline(1024).rstrip("\r\n")
    logger.info("Startup string: %s" % resp)
    magic = resp[4:]
    logger.debug(magic)
    m = hashlib.md5()

    m.update(magic + "password")
    dtnin_digest = m.hexdigest()

    rm("CAPA")
    r("APOP elwynd " + dtnin_digest)
    r("USER elwynd")
    r("PASS password")
    r("STAT")
    rm("LIST")
    r("LIST 2")
    r("LIST 245")
    r("LAST")
    r("RETR")
    rm("RETR 2")
    rm("RETR 345")
    r("LAST")
    rm("UIDL")
    r("UIDL 5")
    r("UIDL -2")
    rm("TOP 3 2")
    rm("TOP")
    rm("TOP 34")
    rm("TOP 3 a")
    rm("TOP b 5")
    rm("TOP 3 -1")
    r("DELE")
    r("DELE 675")
    r("STAT")
    r("DELE 2")
    r("STAT")
    r("RSET")
    r("DELE 2")
    r("STAT")
    r("LAST")
    r("QUIT")

    server.end_run()
        

        
