#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


"""
SMTP server intended for use in mail client nodes that will be passing
messages to DTN for conveyance to the gateway or another nomadic node.

The server implements a minimal SMTP (not ESMTP - consider use of DSN later)
protocol server

Received messages have an X-DTN-rcpts header added containing the set of
recipients to which the message is being sent.

If any of the recipients is in the DTN domain, the message is forked.
One copy is sent to each DTN domain dcipient over DTN,and any other
recipients are handled by sending the message to the gateway.

Each copy of the message is stored in a maildir file  - the lines are terminated with \n

Revision History
================
Version   Date       Author         Notes
0.1       14/03/2010 Elwyn Davies   License statement added.
0.0	  25/07/2009 Elwyn Davies   Created for N4C Summer tests 2009
"""

import os
import sys
import time
import socket
from SocketServer2_6 import *
import threading
import logging
import Queue
from select import select

import pypop_maildir
from dp_msg_send_evt import *

__version__ = "SMTP to DTN server - Python - v0.1"

class SmtpBundleFile:
    # Container for mailbox mail object and destination
    def __init__(self, mail_msg, destn):
        # Maildir message instance
        self.msg = mail_msg
        # List of destinations
        self.destn = destn


class SMTPMessageHandler(BaseRequestHandler):
    """.
    client_address is a tuple containing (ipaddr, port) of the client that made the
    socket connection to our smtp port.

    mailfrom is the raw address the client claims the message is coming
    from.

    rcpttos is a list of raw addresses the client wishes to deliver the
    message to.

    data is a string containing the entire full text of the message,
    headers (if supplied) and all.  It has been `de-transparencied'
    according to RFC 821, Section 4.5.2.  In other words, a line
    containing a `.' followed by other text has had the leading dot
    removed.

    This function should return None, for a normal `250 Ok' response;
    otherwise it returns the desired response string in RFC 821 format.

    """
 
    lineseps = "\r\n"
    UNIXFROM = "From "
    UNIXFROM_LEN = len(UNIXFROM)


    # SMTP states
    COMMAND = 0
    DATA = 1

    def setup(self):
        self.connection = self.request
        self.wfile = self.connection.makefile('wb', 0)

    def finish(self):
        if not self.wfile.closed:
            self.wfile.flush()
        self.wfile.close()
            

    def close_smtp_handler(self):
        self.smtp_run = False

    def handle(self):
        self.smtp_thread = threading.currentThread()
        self.smtp_thread.setName("SMTP server - %d" %
                                 self.server.next_server_num)
        self.server.next_server_num += 1

        # Record this thread with the server
        self.server.add_thread(self)

        self.smtp_run= True

        self.loginfo = self.server.logger.info
        self.logdebug = self.server.logger.debug
        self.logwarn = self.server.logger.warn
        self.logerror = self.server.logger.error

        self.fqdn = socket.getfqdn()
        conn_peer = self.request.getpeername()

        self.logdebug("New SMTP connection from %s" % self.client_address[0])

        # Only accept connection from localhost
        if self.client_address[0] == "127.0.0.1":
            # Write greeting
            self.logdebug("Greeting sent %s %s" % (self.fqdn, __version__))
            self.wfile.write('220 %s %s\r\n' % (self.fqdn, __version__))
        else:
            # Repulse overtures
            self.loginfo("Connection from %s refused" % self.client_address[0])
            self.wfile.write('554 Error: Connection Refused\r\n')
            self.request.close()
            return
            
        # Loop processing incoming messages
        while (self.smtp_run):

            # Setup handler state
            self.state = self.COMMAND
            self.greeting = 0
            self.mailfrom = None
            self.rcpttos = []
            msg_data = ''
            self.msgs = []
            first_data = True
            receive_fd = self.request.fileno()
            buf =""

            # Loop processing lines within a message
            try:
                while(self.smtp_run):
                    i = buf.find('\n')
                    if i == -1:
                       # Read some more
                        # Drop out periodically to check we haven't been stopped
                        rd_fd, wr_fd, err_fd = select([receive_fd], [], [receive_fd], 5)
                        if (len(err_fd) > 0) and (err_fd[0] == receive_fd):
                            self.logdebug("Socket closed")
                            self.smtp_run = False
                            continue
                        if (len(rd_fd) != 1) or (rd_fd[0] != receive_fd):
                            if len(rd_fd) != 0:
                                self.logerror("Spurious output from select")
                            # Loop round again
                            continue
                        more = self.request.recv(1024)
                        if more == "":
                            self.logdebug("EOF occurred")
                            self.smtp_run = False
                            continue
                        buf += more
                        i = buf.find('\n')
                        if i == -1:
                            self.logdebug("Stuck with partial line")
                            continue
                    if i == len(buf) - 1:
                        msg_data = buf
                        buf = ""
                    else:
                        msg_data = buf[:i+1]
                        buf = buf[i+1:]
                        
                    # Analyse incoming message
                    msg_data = msg_data.rstrip(self.lineseps)
                    self.logdebug("In state %d Line received: %s" % (self.state, msg_data))
                    if self.state == self.COMMAND:
                        if msg_data == "":
                            self.wfile.write("500 Error: bad syntax\r\n")
                            break
                        i = msg_data.find(" ")
                        if i < 0:
                            command = msg_data.upper()
                            arg = None
                        else:
                            command = msg_data[:i].upper()
                            arg = msg_data[i+1:].strip()
                        method = getattr(self, 'smtp_' + command, None)
                        if not method:
                            self.wfile.write("502 Error: command '%s' not implemented\r\n" % command)
                            break
                        self.logdebug("Executing command: %s %s" % (command, arg))
                        method(arg)
                    elif self.state == self.DATA:
                        # We will have a list of messages under construction
                        # Write each line to each mess until there is a line with just a "." 
                        if msg_data != ".":
                            # Remove 'quoting' for lines starting with '.'
                            if msg_data[:2] == "..":
                                msg_data = msg_data[1:]
                            for m in self.msgs:
                                # self.logdebug("Writing to %s %s" % (m.msg, " ".join(m.destn)))
                                if first_data:
                                    # self.logdebug("0")
                                    if msg_data[:self.UNIXFROM_LEN] != self.UNIXFROM:
                                        # Synthesize FROM envelope header
                                        # self.logdebug("1")
                                        # self.logdebug("%s%s %s" % (self.UNIXFROM, self.mailfrom, time.asctime()))
                                        m.msg.lineReceived("%s%s %s" % (self.UNIXFROM, self.mailfrom, time.asctime()))
                                        # Output the received data
                                        # self.logdebug("2")
                                        m.msg.lineReceived(msg_data)
                                    else:
                                        # Reuse existing envelope header
                                        # self.logdebug("3")
                                        m.msg.lineReceived(msg_data)
                                else:
                                    # self.logdebug("4")
                                    m.msg.lineReceived(msg_data)
                            first_data = False
                            # self.logdebug("Next line")
                        else:
                            # Finalize all the messages
                            # self.logdebug("5")
                            for m in self.msgs:
                                if m.msg.eomReceived():
                                    fn = m.msg.getFinalName()
                                    self.logdebug("Finishing message for %s in %s" % (" ".join(m.destn), fn))
                                    try:
                                        evt = msg_send_evt(MSG_DTN, fn, MSG_NEW, m.destn)
                                        self.server.send_queue.put_nowait(evt)
                                    except QueueFull:
                                        # The periodic thread will pick up new messages later
                                        pass
                                    finally:
                                        pass
                            # Finished this message - get ready for next one
                            self.msgs = []
                            self.wfile.write('250 Ok\r\n')
                            self.logdebug("Finished message")
                            break
                    else:
                        self.logerror("SMTP state has illegal value: %d" % self.state)
                        self.wfile.write("451 Internal confusion - bad state value\r\n")
                        self.smtp_run = False
                        self.request.close()
                        break
                # End of while loop - ready for next command or connection close
            except:
                self.logdebug("Exception %s" % sys.exc_info()[0])
                print sys.exc_info()
                if not self.smtp_run:
                    self.request.close()
            finally:
                if len(self.msgs) > 0:
                    for m in self.msgs:
                        m.msg.connectionLost()
            self.logdebug("Next iteration")
        self.server.remove_thread(self)
        self.finish()
        return

    # SMTP and ESMTP commands
    def smtp_HELO(self, arg):
        if not arg:
            self.wfile.write('501 Syntax: HELO hostname\r\n')
            return
        if self.greeting:
            self.wfile.write('503 Duplicate HELO/EHLO\r\n')
        else:
            self.greeting = arg
            self.wfile.write('250 %s\r\n' % self.fqdn)

    def smtp_NOOP(self, arg):
        if arg:
            self.wfile.write('501 Syntax: NOOP\r\n')
        else:
            self.wfile.write('250 Ok\r\n')

    def smtp_QUIT(self, arg):
        # args is ignored
        self.wfile.write('221 Bye\r\n')
        self.wfile.flush()
        self.request.close()
        self.smtp_run = False

    # used for both MAIL and RCPT
    def __getaddr(self, keyword, arg):
        address = None
        keylen = len(keyword)
        if arg[:keylen].upper() == keyword:
            address = arg[keylen:].strip()
            if not address:
                pass
            elif address[0] == '<' and address[-1] == '>' and address != '<>':
                # Addresses can be in the form <person@dom.com> but watch out
                # for null address, e.g. <>
                address = address[1:-1]
        self.logdebug("__getaddr returning %s" % address)
        return address

    def smtp_MAIL(self, arg):
        self.logdebug('===> MAIL %s' % arg)
        address = self.__getaddr('FROM:', arg)
        if not address:
            self.wfile.write('501 Syntax: MAIL FROM:<address>\r\n')
            return
        if self.mailfrom:
            self.wfile.write('503 Error: nested MAIL command\r\n')
            return
        self.mailfrom = address
        self.logdebug('sender: %s' % self.mailfrom)
        self.wfile.write('250 Ok\r\n')

    def smtp_RCPT(self, arg):
        self.logdebug('===> RCPT %s ' % arg)
        if not self.mailfrom:
            self.wfile.write('503 Error: need MAIL command\r\n')
            return
        address = self.__getaddr('TO:', arg) if arg else None
        if not address:
            self.wfile.write('501 Syntax: RCPT TO: <address>\r\n')
            return
        self.rcpttos.append(address)
        self.logdebug('recips: %s', " ".join(self.rcpttos))
        self.wfile.write('250 Ok\r\n')

    def smtp_RSET(self, arg):
        if arg:
            self.wfile.write('501 Syntax: RSET\r\n')
            return
        # Resets the sender, recipients, and data, but not the greeting
        self.mailfrom = None
        self.rcpttos = []
        self.state = self.COMMAND
        self.wfile.write('250 Ok\r\n')

    def smtp_DATA(self, arg):
        if not self.rcpttos:
            self.wfile.write('503 Error: need RCPT command\r\n')
            return
        if arg:
            self.wfile.write('501 Syntax: DATA\r\n')
            return
        self.state = self.DATA
        self.wfile.write('354 End data with <CR><LF>.<CR><LF>\r\n')

        # Create message object(s) to write data into
        gateway_rcpts = []
        for r in self.rcpttos:
            if r.endswith("@"+self.server.domain):
                # This should be split off and sent in a separate bundle
                self.loginfo("Bundle to %s" % r)
                msg = self.server.mailbox.newMessage(r)
                self.msgs.append(SmtpBundleFile(msg, [r]))
            else:
                gateway_rcpts.append(r)
        if gateway_rcpts != []:
            self.loginfo("Common bundle to %s" % " ".join(gateway_rcpts))
            msg = self.server.mailbox.newMessage( " ".join(gateway_rcpts))
            self.msgs.append(SmtpBundleFile(msg, gateway_rcpts))           
        

class DtnSmtpServer(ThreadingMixIn, TCPServer):

    def __init__(self, addr, handler, domain, mailbox, send_queue, logger):
        # These are used  by individual requests
        # accessed via self.server in the handle function
        self.domain = domain
        self.mailbox = mailbox
        self.send_queue = send_queue
        self.logger = logger
        self.running_threads = set()
        self.next_server_num = 1

        TCPServer.__init__(self, addr, handler,
                           bind_and_activate = False)
        self.allow_reuse_address = True
        self.server_bind()
        self.server_activate()

        self.daemon_threads = True
        
    def add_thread(self, thread):
        self.logger.debug("New thread added")
        self.running_threads.add(thread)

    def remove_thread(self, thread):
        if thread in self.running_threads:
            self.running_threads.remove(thread)

    def end_run(self):
        for thread in self.running_threads:
            if thread.smtp_thread.isAlive():
                thread.close_smtp_handler()
        del self.running_threads
        self.shutdown()

def dtn_smtp_server(domain, mailbox, send_queue, logger):
    HOST, PORT = "localhost", 2125
    return DtnSmtpServer((HOST, PORT), SMTPMessageHandler,
                          domain, mailbox, send_queue, logger)


if __name__ == "__main__":

    import smtplib
    import email

    # Basic test client
    def client(ip, port, message):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((ip, port))
        sock.send(message)
        response = sock.recv(1024)
        print "Received: %s" % response
        sock.close()

    # Port 0 means to select an arbitrary unused port
    HOST, PORT = "localhost", 0

    logger = logging.getLogger("test")
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    fmt = logging.Formatter("%(levelname)s %(threadName)s %(message)s")
    ch.setFormatter(fmt)
    logger.addHandler(ch)

    # Create an SMTP client
    smtp_client = smtplib.SMTP()

    # Create the send queue
    send_q = Queue.Queue()

    nomadic_domain = "nomadic.n4c.eu"
    tmp_domain = pypop_maildir.MaildirDirdbmDomain("/tmp")
    if not tmp_domain.userDirectory("dtnout"):
        tmp_domain.addUser("dtnout", "password")
    dtnout_mailbox = pypop_maildir.MaildirMailbox(tmp_domain.userDirectory("dtnout"),
                                                 "dtnout")
    
    server = DtnSmtpServer((HOST, PORT), SMTPMessageHandler,
                           nomadic_domain, dtnout_mailbox, send_q, logger)
    ip, port = server.server_address

    # Start a thread with the server -- that thread will then start one
    # more thread for each request
    server_thread = threading.Thread(target=server.serve_forever)
    # Exit the server thread when the main thread terminates
    server_thread.setDaemon(True)
    server_thread.start()
    logger.info( "Server loop running in thread:%s" % server_thread.getName())

    # Basic startup tests
    client(ip, port, "Hello World 1")
    client(ip, port, "Hello World 2")

    # Read in test messages
    fn = open("msg3")
    msg3 = email.message_from_file(fn)

    # SMTP tests
    resp = smtp_client.connect(ip, port)
    print resp
    logger.info(("SMTP connect replied: %s %s" % resp))
    smtp_client.helo("Tester on localhost")
    logger.info("SMTP greeted with: %s" %smtp_client.helo_resp )
    logger.info("sending msg3 with 1 recipient")
    smtp_client.sendmail("test@nomadic.n4c.eu", "mumble@example.com", msg3.as_string())
    logger.info("sending msg3 with 3 recipients")
    smtp_client.sendmail("test@nomadic.n4c.eu",
                         ["mumble@example.com", "jackass@nomadic.n4c.eu", "a@b.c"],
                         msg3.as_string(True))

    logger.info("Testing RSET after MAIL")
    resp = smtp_client.mail("a@b.c")
    resp = smtp_client.rset()
    logger.info("Expecting '250 Ok' response from RSET: got %s %s" % resp)
    logger.info("Testing RSET after MAIL and RCPT")
    smtp_client.mail("a@b.c")
    smtp_client.rcpt("d@e.f")
    resp = smtp_client.rset()
    logger.info("Expecting '250 Ok' response from RSET: got %s %s" % resp)
    logger.info("Testing RCPT without preceding MAIL")
    resp = smtp_client.rcpt("d@e.f")
    logger.info("Expecting '503 Error: need MAIL command' response from RCPT: got %s %s" % resp)
    smtp_client.rset()
    logger.info("Closing connection with QUIT")
    resp = smtp_client.quit()
    logger.info("SMTP quit replied (nothing if running in Python 2.5): %s" %resp)

    while not send_q.empty():
        evt = send_q.get()
        logger.info("Msg to %s" % evt.destination())
    server.end_run()
