#! /usr/bin/python
# PyMail DTN Nomadic Mail System
# Copyright (C) Folly Consulting Ltd, 2009, 2010
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
# 
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


"""
DTN NAT relay linkage

This program provides half the relay between the static routing domain used
by TCD's 'village' email relay and the nomadic routing domain that uses PRoPHET
routing.

It registers to receive any of the emails or reports that might come in to the
relay in the static routing domain and sends them on to the PRoPHET domain.

This requires destination address translation!

To minimize the routing in the static domain we use a single eid and multiplex
different users onto it.  These are then sent on to to the different eids in the
PRoPHET domain.
Thus
dtn://relay.nomadic.n4c.eu/<user>/email/in
is mapped to
dtn://<user>.nomadic.n4c.eu/email/in

We also receive the delivery reports from the gateway but we will just junk them
for the time being (or maybe not bother with delivery reports at all on the static
side for the time being.)

  """

import os
import time
import threading
import dtnapi
import socket
import Queue
import logging
import logging.handlers
import logging.config
import re
from select import select

# Expiration period for mail bundles
# Try 3 days for now - value in seconds
MAIL_EXPIRY = (3 * 24 * 60 * 60)

# Status report flag values
# See 
REPORT_STATUS_DELIVERED = 8
REPORT_STATUS_DELETED = 16

# API Ports used for ProPHET and Static regions
# Declared as strings so they can be pushed into the env variable value...
PRoPHET_API_PORT_STR = "5011"
STATIC_API_PORT_STR = "5010"

# To connect to the right daemon we have to set the DTNAPI_PORT envvironment
# variable before call dtn_open
API_PORT_VAR = "DTNAPI_PORT"

# List of users we are looking after
users = ["elwynd", "user1", "user2", "user3", "user4", "user5", "user6", "user7", "user8", "user9", "user10"]

# Local part of source/destination EID
EMAIL_OUT = "email/out"
EMAIL_IN = "email/in"

# Local part of report EID
EMAIL_REPORTS = "email/reports"


# Function to make the EIDs
def build_eids(dtn_handle):
    if dtn_handle == -1:
        raise DtnError("bad DTN handle")
    email_out_addr = dtnapi.dtn_build_local_eid(dtn_handle, EMAIL_OUT)
    email_in_addr = dtnapi.dtn_build_local_eid(dtn_handle, EMAIL_IN)
    report_addr = dtnapi.dtn_build_local_eid(dtn_handle, EMAIL_REPORTS)
    if email_in_addr == None or email_out_addr == None or report_addr == None:
        raise DtnError("failure in dtn_build_local_eid")
    return (email_out_addr, email_in_addr, report_addr)

def main(log_file, log_config_file):
    # Keep the loop running
    relay_in_run = True

    # Setup logging
    logging.config.fileConfig(log_config_file)
    pymail_logger = logging.getLogger("dtn_pymail")
    loginfo = pymail_logger.info
    logerror = pymail_logger.error
    logdebug = pymail_logger.debug

    # Set domain for which we are handling mail
    mail_domain = "nomadic.n4c.eu"

    # RE pattern for converting destination addresses
    nat_patt = re.compile(r"dtn://relay\.nomadic\.n4c\.eu/([^/]*)/email/in$")

    # Create a connection to the PRoPHET side DTN daemon
    os.putenv(API_PORT_VAR, PRoPHET_API_PORT_STR)
    prophet_dtn_handle = dtnapi.dtn_open()
    if prophet_dtn_handle == -1:
        logerror("dp_relay_in: unable to open connection with PRoPHET daemon")
    prophet_eid = dtnapi.dtn_build_local_eid(prophet_dtn_handle, "")
    loginfo("dp_relay_in: Connected to PRoPHET daemon at %s", prophet_eid)

    # Generate the (outgoing) EIDs
    (email_out_eid, email_in_eid, report_eid) = build_eids(prophet_dtn_handle)

    # Create a connection to the Static side DTN daemon
    os.putenv(API_PORT_VAR, STATIC_API_PORT_STR)
    static_dtn_handle = dtnapi.dtn_open()
    if static_dtn_handle == -1:
        logerror("dp_relay_in: unable to open connection with Static daemon")
    static_eid = dtnapi.dtn_build_local_eid(static_dtn_handle, "")
    static_report_address = dtnapi.dtn_build_local_eid(static_dtn_handle,
                                                       EMAIL_REPORTS)
    loginfo("dp_relay_in: Connected to Static daemon at %s", static_eid)

    # Destination address for 

    # Prepare for sending bundles on.
    # Specify a basic bundle spec
    # - We always send the payload as a permanent file
    send_pt = dtnapi.DTN_PAYLOAD_FILE
    # - The source address is always the email_addr
    # - The report address is always the report_addr
    # - The registration id needed in dtn_send is the place
    #   where reports come back to.. we always want reports
    #   but it is unclear why we need to subscribe to the 'session'
    #   just to do the send.The reports will come back through
    #   another connection. Lets try with no regid
    send_regid = dtnapi.DTN_REGID_NONE
    # - The destination address has to be synthesized (later)
    # - We want delivery reports (and maybe deletion reports?)
    send_dopts = dtnapi.DOPTS_DELIVERY_RCPT
    # - Send with normal priority.
    send_pri = dtnapi.COS_NORMAL
    # Mail bundlesshould last a while..
    send_exp = MAIL_EXPIRY


    # Register to receive bundles on email in addresses for each user
    for user in users:
        # Address at which email will arrive on static relay
        user_email_in_addr = static_eid + user + "/" + EMAIL_IN
        
        # Check if email/in registration exists and register if not
        # Otherwise bind to the existing registration
        regid = dtnapi.dtn_find_registration(static_dtn_handle,
                                             user_email_in_addr)
        if (regid == -1):
            # Need to register the EID.. make it permanent with 'DEFER'
            # characteristics so that bundles are saved if theye arrive
            # while the handler is inactive
            # Expire the registration a long time in the future
            exp = 365 * 24 * 60 * 60
            # The registration is immediately active
            passive = False
            # We don't want to execute a script
            script = ""
            
            regid = dtnapi.dtn_register(static_dtn_handle, user_email_in_addr,
                                        dtnapi.DTN_REG_DEFER,
                                        exp, passive, script)
        else:
            dtnapi.dtn_bind(static_dtn_handle, regid)
        
    # Still on the Static side...
    # Check if email/report registration exists and register if not
    static_report_addr = dtnapi.dtn_build_local_eid(static_dtn_handle,
                                                    EMAIL_REPORTS)
    # Otherwise bind to the existing registration
    regid = dtnapi.dtn_find_registration(static_dtn_handle, static_report_addr)
    if (regid == -1):
        # Need to register the EID.. make it permanent with 'DEFER'
        # characteristics so that bundles are saved if theye arrive
        # while the handler is inactive
        # Expire the registration a long time in the future
        exp = 365 * 24 * 60 * 60
        # The registration is immediately active
        passive = False
        # We don't want to execute a script
        script = ""
        
        regid = dtnapi.dtn_register(static_dtn_handle, static_report_addr,
                                    dtnapi.DTN_REG_DEFER,
                                    exp, passive, script)
    else:
        dtnapi.dtn_bind(static_dtn_handle, regid)

    logdebug("dp_relay_in: Entering DTN relay incoming loop")

    # Process mail
    # Now sit and wait for incoming email
    # Note that just using dtn_recv with a timeout doesn't work.
    # The blocking I/O upsets the threading seemingly.
    receive_fd = dtnapi.dtn_poll_fd(static_dtn_handle)
    while (relay_in_run):
        # Poll currently sets timeout in ms - this is a bug
        # but -1 means infinity
        dtnapi.dtn_begin_poll(static_dtn_handle, -1)
        # Wait until some IO comes in.
        # This is either a received bundle or an Interrupt (SIGINT).
        try:
            rd_fd, wr_fd, err_fd = select([receive_fd], [], [])
        except:
            # The system call was interrupted
            dtnapi.dtn_cancel_poll(static_dtn_handle)
            relay_in_run = False
            break

        # The fd is allegedly readable
        # Cancel the poll anyway
        dtnapi.dtn_cancel_poll(static_dtn_handle)

        # If there are no readable file desciptors something went wrong
        if (len(rd_fd) == 0):
            logerror("dp_relay_in: No readable filedescriptor returned from select")
            relay_in_run = False
            break

        # Check the dtn fd was actually readable
        if not (receive_fd in rd_fd):
            # this is an error
            self.logerror("dp_relay_in: Unknown fd has become readbale")
            self.receive_run = False
            break
        
        """
        There should always be something to read
        Put in a timeout just in case
        The call to dtn_recv terminates the poll
        We accept the email as a (temporary) file.
        There is a nasty snag here. There is small window
        of opportunity where the bundle has been written
        to a temporary file delivered to this application
        but before file has been renamed and registered in the maildir
        and database. If the application exits during this period
        the email will be lost.  This is not a happy situation.
        At present DTN2 does not have a means for dtn_recv to offer
        a filename that could be used.
        
        Uing DTN_PAYLOAD_MEM would present a similar problem
        During the period the memory image is being written to file
        in the application there is no permanent  copy of the file
        but the DTN daemon believes it has delivered and is busy sending
        a delivery notification.

        NOTE: On receiving the file, the file name is in the bundle
        payload as a NULL terminated string.  Python leaves the terminating
        byte in place.
        """
        email = dtnapi.dtn_recv(static_dtn_handle, dtnapi.DTN_PAYLOAD_FILE, 1)
        if email == None:
            # just go back to listening (xxxxebd is this sensible?)
            logerror("dp_relay_in: No readable bundle received after select")
            continue

        # Check for delivery reports
        if email.dest == static_report_address:
            # For the time being log the report and continue..
            # later we will redirect these as well
            if email.status_report != None:
                logdebug("dp_relay_in: Received status report")
                if email.status_report.flags == REPORT_STATUS_DELIVERED:
                    loginfo( "dp_relay_in: Received delivery report re from %s sent %d seq %d" %
                             (email.status_report.bundle_id.source,
                              email.status_report.bundle_id.creation_secs,
                              email.status_report.bundle_id.creation_seqno))
                    # Update the database
  
                elif email.status_report.flags == REPORT_STATUS_DELETED:
                    loginfo("dp_relay_in: Received deletion report re from %s sent %d seq %d" %
                            (email.status_report.bundle_id.source,
                             email.status_report.bundle_id.creation_secs,
                             email.status_report.bundle_id.creation_seqno))
                    # Update the database

                else:
                    logwarn("dp_relay_in: Received unexpected report: Flags: %d" % email.status_report.flags)

            continue

        # Modify the destination address and resend into PRoPHET domain
        fn = email.payload
        l = len(fn)
        if fn[l-1] == "\x00":
            fn = fn[:-1]
        logdebug("dp_relay_in: Got incoming bundle in file %s" % fn)
        loginfo("dp_relay_in: Received email bundle from %s sent %d seq %d" %
                (email.source,
                 email.creation_secs,
                 email.creation_seqno))
        
        # Build the destination EID
        nat_match = nat_patt.match(email.dest)
        if nat_match == None:
            logerror("dp_relay_in: Non-standard bundle destination %s" % email.dest)
            continue
        
        destn_eid = "dtn://%s.%s/%s" % (nat_match.group(1), mail_domain, EMAIL_IN)

        loginfo("dp_relay_in: Sending message to %s" % destn_eid)

        # Send the bundle
        # The last two empty strings are to do with sessions which we don't use
        bundle_id = dtnapi.dtn_send(prophet_dtn_handle, send_regid,
                                    email_out_eid, destn_eid,
                                    report_eid, send_pri, send_dopts,
                                    send_exp, send_pt, fn, "", "")
        if bundle_id == None:
            logwarn("dp_relay_in: Sending of message to %s failed" % destn_eid)
        else:
            # Store the details of the sent bundle
            loginfo("dp_relay_in: %s sent at %d, seq no %d" %
                    (bundle_id.source,
                     bundle_id.creation_secs,
                     bundle_id.creation_seqno))

    dtnapi.dtn_close(static_dtn_handle)
    dtnapi.dtn_close(prophet_dtn_handle)
    loginfo("dp_relay_in exiting")

if __name__ == "__main__":
    main("/home/user/dtnrun/logs", "/home/user/dtnrun/python/dtn_pymail_log.conf")

